#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains the class definition for the MacroServer generic
scan"""

from __future__ import with_statement
from __future__ import print_function

__all__ = ["OverloadPrint", "PauseEvent", "Hookable", "ExecMacroHook",
           "MacroFinder", "Macro", "macro", "iMacro", "imacro",
           "MacroFunc", "Type", "ParamRepeat", "Table", "List", "ViewOption",
           "LibraryError"]

__docformat__ = 'restructuredtext'

import threading
import traceback
import time
import sys
import operator
import types
import weakref
import StringIO
import ctypes

from taurus.core.util.log import Logger
from taurus.core.util.prop import propertx
from taurus.console.table import Table
from taurus.console.list import List

from taurus.core.tango.sardana.pool import PoolElement

from sardana.sardanadefs import State
from sardana.util.wrap import wraps

from .msparameter import Type, ParamType, ParamRepeat
from .msexception import StopException, AbortException, \
    MacroWrongParameterType, UnknownEnv, UnknownMacro, LibraryError
from .msoptions import ViewOption

asyncexc = ctypes.pythonapi.PyThreadState_SetAsyncExc
# first define the async exception function args. This is
# absolutely necessary for 64 bits machines.
asyncexc.argtypes = (ctypes.c_long, ctypes.py_object)


class OverloadPrint(object):

    def __init__(self, m):
        self._macro = m
        self._accum = ""

    def __enter__(self):
        self.stdout = sys.stdout
        sys.stdout = self

    def __exit__(self, exc_type, exc_value, traceback):
        self.flush()
        sys.stdout = self.stdout

    def write(self, s):
        self._accum += s
        # while there is no new line, just accumulate the buffer
        try:
            if s[-1] == '\n' or s.index('\n') >= 0:
                self.flush()
        except ValueError:
            pass

    def flush(self):
        b = self._accum
        if b is None or len(b) == 0:
            return
        #take the '\n' because the output is a list of strings, each to be
        #interpreted as a separate line in the client
        if b[-1] == '\n': b = b[:-1]
        self._macro.output(b)
        self._accum = ""


class PauseEvent(Logger):

    def __init__(self, macro_obj, abort_timeout = 0.2):
        self._name = self.__class__.__name__
        self._pause_cb = None
        self._resume_cb = None
        self._macro_obj_wr = weakref.ref(macro_obj)
        self._macro_name = macro_obj._getName()
        self._wait_for_abort_exception = False
        self._wait_for_abort_timeout = abort_timeout
        Logger.__init__(self, "Macro_%s %s" % (self._macro_name, self._name))
        # we create an event object that is automatically set
        self._event = threading.Event()
        self._event.set()

    @property
    def macro_obj(self):
        return self._macro_obj_wr()

    def pause(self, cb=None):
        self.debug("[START] Pause")
        self._pause_cb = cb
        self._event.clear()
        self.debug("[ END ] Pause")

    def resume(self, cb=None):
        if self.isPaused():
            self.debug("[START] Resume")
            self._resume_cb = cb
            self._event.set()
            self.debug("[ END ] Resume")

    def resumeForAbort(self):
        if self.isPaused():
            self.debug("[RESUME] (Abort)")
            self._wait_for_abort_exception = True
            self._event.set()

    def wait(self,timeout=None):
        pauseit = not self._event.isSet()
        if pauseit and self._pause_cb is not None:
            self._pause_cb(self.macro_obj)
        self._event.wait(timeout)
        # if an event is set because an abort has been issued during a paused
        # macro wait for the ashyncronous AbortException to arrive at this thread
        if self._wait_for_abort_exception:
            self._wait_for_abort_exception = False
            time.sleep(self._wait_for_abort_timeout)
            self.debug('Abort exception did not occured in pause for %ss.' \
                       'Performing a Forced Abort.' % self._wait_for_abort_timeout)
            raise AbortException("Forced")
        if pauseit and self._resume_cb is not None:
            self._resume_cb(self.macro_obj)

    def isPaused(self):
        return not self._event.isSet()


class Hookable(Logger):

    # avoid creating an __init__

    def _getHooks(self):
        try:
            return self._hooks
        except:
            self._hooks = []
        return self._hooks

    def _getHookHintsDict(self):
        try:
            return self._hookHintsDict
        except:
            self._hookHintsDict = {'_ALL_':[], '_NOHINTS_':[]}
        return self._hookHintsDict

    def getAllowedHookHints(self):
        return self.__class__.hints.get('allowsHooks')

    def getHints(self):
        return self._getHookHintsDict().keys()

    def getHooks(self, hint=None):
        '''This will return a list of hooks that have the given hint. Two reserved
        hints are always valid:
          - "_ALL_": which contains all the hooks
          - "_NOHINTS_": which contains the hooks that don't provide any hint

        :param hint: (str) a hint. If None is passed, it returns a list of
                     (hook,hints) tuples

        :return: (list) an ordered list of hooks that have the given hint
        '''
        if hint is None:
            return self._getHooks()
        else:
            return self._getHookHintsDict().get(hint,[])

    @propertx
    def hooks():
        def get(self):
            return self._getHooks()

        def set(self, hooks):
            '''hooks must be list<callable,list<str>>. Exceptionally, for
            backwards compatibility, list<callable> is also admitted, but may
            not be supported in the future.
            "two variables are created:
                - self._hooks (list<callable,list<str>>) (will be a tuple
                              regardless of what was passed)
                - self._hookHintsDict (dict<str,list>) a dict of key=hint and
                                      value=list of hooks with that hint.
                                      self._hookHintsDict also stores two
                                      special keys: "_ALL_": which contains all
                                      the hooks "_NOHINTS_": which contains the
                                      hooks that don't provide hints
            '''
            if not isinstance(hooks, list):
                self.error('the hooks must be passed as a list<callable,list<str>>')
                return

            #store self._hooks, making sure it is of type: list<callable,list<str>>
            self._hooks = []
            for h in hooks:
                if  isinstance(h,(tuple, list)) and len(h)==2:
                    self._hooks.append(h)
                else: #we assume that hooks is a list<callable>
                    self._hooks.append((h,[]))
                    self.info('Deprecation warning: hooks should be set with a list of hints. See Hookable API docs')

            #create _hookHintsDict
            self._getHookHintsDict()['_ALL_'] = zip(*self._hooks)[0]
            nohints = self._hookHintsDict['_NOHINTS_']
            for hook,hints in self._hooks:
                if len(hints) == 0:
                    nohints.append(hook)
                else:
                    for hint in hints:
                        try:
                            self._hookHintsDict[hint].append(hook)
                        except KeyError:
                            self._hookHintsDict[hint] = [hook]
        return get,set


class ExecMacroHook(object):
    """A speciallized callable hook for executing a sub macro inside another
    macro as a hook"""

    def __init__(self, parent_macro, *pars, **kwargs):
        self._macro_obj_wr = weakref.ref(parent_macro)
        self._pars = pars
        self._opts = kwargs

    @property
    def macro_obj(self):
        return self._macro_obj_wr()

    def __call__(self):
        self.macro_obj.execMacro(*self._pars, **self._opts)


class MacroFinder:

    def __init__(self, macro_obj):
        self._macro_obj_wr = weakref.ref(macro_obj)

    @property
    def macro_obj(self):
        return self._macro_obj_wr()

    def __getattr__(self, name):

        def f(*args, **kwargs):
            p_m = self.macro_obj
            p_m.syncLog()
            opts = { 'parent_macro' : p_m,
                     'executor'     : p_m.executor }
            kwargs.update(opts)
            eargs = [name]
            eargs.extend(args)
            return p_m.execMacro(*eargs, **kwargs)

        setattr(self, name, f)

        return f

def mAPI(fn):
    """Wraps the given Macro method as being protected by the stop procedure.
    To be used by the :class:`Macro` as a decorator for all methods.
    :param: macro method
    :return: wrapped macro method"""
    @wraps(fn)
    def new_fn(*args, **kwargs):
        self = args[0]
        if not self.isProcessingStop():
            is_macro_th = self._macro_thread == threading.current_thread()
            if self._shouldRaiseStopException():
                if is_macro_th:
                    self.setProcessingStop(True)
                raise StopException("stopped before calling %s" % fn.__name__)
        ret = fn(*args, **kwargs)
        if not self.isProcessingStop():
            if self._shouldRaiseStopException():
                if is_macro_th:
                    self.setProcessingStop(True)
                raise StopException("stopped after calling %s" % fn.__name__)
        return ret
    return new_fn


class macro(object):
    """Class designed to decorate a python function to transform it into a
    macro. Examples::

        @macro()
        def my_macro1(self):
            self.output("Executing %s", self.getName())

        @macro([ ["moveable", Type.Moveable, None, "motor to watch"] ])
        def where_moveable(self, moveable):
            self.output("Moveable %s is at %s", moveable.getName(), moveable.getPosition())"""

    def __init__(self, param_def=None, result_def=None, env=None, hints=None,
                 interactive=None):
        self.param_def = param_def
        self.result_def = result_def
        self.env = env
        self.hints = hints
        self.interactive = interactive

    def __call__(self, fn):
        fn.macro_data = {}
        fn.param_def = self.param_def
        fn.result_def = self.result_def
        fn.hints = self.env
        fn.env = self.hints
        fn.interactive = self.interactive
        return fn


from functools import partial
imacro = partial(macro, interactive=True)


class Macro(Logger):
    """ The Macro base class. All macros should inherit directly or indirectly
    from this class."""

    #: internal variable
    Init     = State.Init

    #: internal variable
    Running  = State.Running

    #: internal variable
    Pause    = State.Standby

    #: internal variable
    Stop     = State.Standby

    #: internal variable
    Fault    = State.Fault

    #: internal variable
    Finished = State.On

    #: internal variable
    Ready    = State.On

    #: internal variable
    Abort    = State.Alarm

    #: Constant used to specify all elements in a parameter
    All      = ParamType.All

    #: internal variable
    BlockStart = '<BLOCK>'

    #: internal variable
    BlockFinish = '</BLOCK>'

    #: This property holds the macro parameter description.
    #: It consists of a sequence of parameter information objects.
    #: A parameter information object is either:
    #:
    #:    #. a simple parameter object
    #:    #. a parameter repetition object
    #:
    #: A simple parameter object is a sequence of:
    #:
    #:    #. a string representing the parameter name
    #:    #. a member of :obj:`Macro.Type` representing the parameter data type
    #:    #. a default value for the parameter or None if there is no default value
    #:    #. a string with the parameter description
    #:
    #: Example::
    #:
    #:     param_def = ( ('value', Type.Float, None, 'a float parameter' ) )
    #:
    #: A parameter repetition object is a sequence of:
    #:
    #:    #. a string representing the parameter repetition name
    #:    #. a sequence of parameter information objects
    #:    #. a dictionary representing the parameter repetition semantics or None
    #:       to use the default parameter repetition semantics. Dictionary keys are:
    #:
    #:           * *min* - integer representing minimum number of repetitions or None
    #:             for no minimum.
    #:           * *max* - integer representing maximum number of repetitions or None
    #:             for no maximum.
    #:
    #:       Default parameter repetition semantics is ``{ 'min': 1, 'max' : None }``
    #:       (in other words, "at least one repetition" semantics)
    #:
    #: Example::
    #:
    #:     param_def = (
    #:         ( 'motor_list', ( ( 'motor', Type.Motor, None, 'motor name') ), None, 'List of motors')
    #:     )
    param_def = []

    #: This property holds the macro result description.
    #: It a single parameter information object.
    #:
    #: .. seealso:: :obj:`~sardana.macroserver.macro.Macro.param_def`
    result_def = []

    #: Hints to give a client to perform special tasks.
    #: Example: scan macros give hints on the types of hooks they support. A
    #: :term:`GUI` can use this information to allow a scan to have sub-macros
    #: executed as hooks.
    hints = {}

    #: a set of mandatory environment variable names without which your macro
    #: cannot run
    env = ()
    
    #: decide if the macro should be able to receive input from the user
    #: [default: False]. A macro which asks input but has this flag set to False
    #: will print a warning message each time it is executed
    interactive = False

    def __init__(self, *args, **kwargs):
        """Constructor"""
        self._name = kwargs.get('as', self.__class__.__name__)
        self._in_pars = args
        self._out_pars = None
        self._aborted = False
        self._stopped = False
        self._processingStop = False
        self._parent_macro = kwargs.get('parent_macro')
        self._executor = kwargs.get('executor')
        self._macro_line = kwargs.get('macro_line')
        self._interactive_mode = kwargs.get('interactive', True)
        self._macro_thread = None
        self._id = kwargs.get('id')
        self._desc = "Macro '%s'" % self._macro_line
        self._macro_status = { 'id' : self._id,
                               'range' : (0.0, 100.0),
                               'state' : 'start',
                               'step' : 0.0 }
        self._pause_event = PauseEvent(self)
        log_parent = self.parent_macro or self.door
        Logger.__init__(self, "Macro[%s]" % self._name, log_parent)
        self._reserveObjs(args)

    ## @name Official Macro API
    #  This list contains the set of methods that are part of the official macro
    #  API. This means that they can be safely used inside any macro.
    #@{

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Methods to be implemented by the actual macros
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def run(self, *args):
        """**Macro API**. Runs the macro. **Overwrite MANDATORY!** Default implementation
        raises RuntimeError.

        :raises: RuntimeError"""
        raise RuntimeError("Macro %s does not implement run method" % self.getName())

    def prepare(self, *args, **kwargs):
        """**Macro API**. Prepare phase. Overwrite as necessary.
        Default implementation does nothing"""
        pass

    def on_abort(self):
        """**Macro API**. Hook executed when an abort occurs.
        Overwrite as necessary. Default implementation does nothing"""
        pass

    def on_pause(self):
        """**Macro API**. Hook executed when a pause occurs.
        Overwrite as necessary. Default implementation does nothing"""
        pass

    def on_stop(self):
        """**Macro API**. Hook executed when a stop occurs.
        Overwrite as necessary. Default implementation calls
        :meth:`~Macro.on_abort`"""
        return self.on_abort()

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # API
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    @mAPI
    def checkPoint(self):
        """**Macro API**.
        Empty method that just performs a checkpoint. This can be used
        to check for the stop. Usually you won't need to call this method"""
        pass

    @mAPI
    def pausePoint(self, timeout=None):
        """**Macro API**.
        Will establish a pause point where called. If an external source as
        invoked a pause then, when this this method is called, it will be block
        until the external source calls resume.
        You may want to call this method if your macro takes a considerable time
        to execute and you may whish to pause it at some time. Example::

            for i in range(10000):
                time.sleep(0.1)
                self.output("At step %d/10000", i)
                self.pausePoint()

        :param timeout:
            timeout in seconds [default: None, meaning wait forever]
        :type timeout: :obj:`float`"""
        return self._pausePoint(timeout=timeout)

    @property
    def macros(self):
        """**Macro API**.
        An object that contains all macro classes as members. With
        the returning object you can invoke other macros. Example::

            m = self.macros.ascan('th', '0', '90', '10', '2')
            scan_data = m.data"""
        self.checkPoint()
        if not hasattr(self, '_macros'):
            self._macros = MacroFinder(self)
        return self._macros

    @mAPI
    def getMacroStatus(self):
        """**Macro API**.
        Returns the current macro status. Macro status is a :obj:`dict` where
        keys are the strings:

            * *id* - macro ID (internal usage only)
            * *range* - the full progress range of a macro (usually a
              :obj:`tuple` of two numbers (0, 100))
            * *state* - the current macro state, a string which can have values
              *start*, *step*, *stop* and *abort*
            * *step* - the current step in macro. Should be a value inside the
              allowed macro range

        :return: the macro status
        :rtype: :obj:`dict`"""
        return self._macro_status

    @mAPI
    def getName(self):
        """**Macro API**.
        Returns this macro name

        :return: the macro name
        :rtype: :obj:`str`"""
        return self._name

    @mAPI
    def getID(self):
        """**Macro API**.
        Returns this macro id

        :return: the macro id
        :rtype: :obj:`str`"""
        return self._id

    @mAPI
    def getParentMacro(self):
        """**Macro API**.
        Returns the parent macro reference.

        :return: the parent macro reference or None if there is no parent macro
        :rtype: :class:`~sardana.macroserver.macro.Macro`"""
        return self._parent_macro

    @mAPI
    def getDescription(self):
        """**Macro API**.
        Returns a string description of the macro.

        :return: the string description of the macro
        :rtype:  :obj:`str`"""
        return self._desc

    @mAPI
    def getParameters(self):
        """**Macro API**.
        Returns a the macro parameters. It returns a list containning the
        parameters with which the macro was executed

        :return: the macro parameters
        :rtype: :obj:`list`"""
        return self._in_pars

    @mAPI
    def getExecutor(self):
        """**Macro API**.
        Returns the reference to the object that invoked this macro. Usually
        is a MacroExecutor object.

        :return: the reference to the object that invoked this macro
        :rtype: :class:`~sardana.macroserver.macromanager.MacroExecutor`"""
        return self._executor

    @mAPI
    def getDoorObj(self):
        """**Macro API**.
        Returns the reference to the Door that invoked this macro.

        :return: the reference to the Door that invoked this macro.
        :rype: :class:`~sardana.macroserver.door.Door`"""
        return self.executor.getDoor()

    @mAPI
    def getManager(self):
        """**Macro API**.
        Returns the manager for this macro (usually a MacroServer)

        :return: the MacroServer
        :rtype: :class:`~sardana.macroserver.macroserver.MacroServer`"""
        return self.door.manager

    manager = property(getManager)

    @mAPI
    def getMacroServer(self):
        """**Macro API**.
        Returns the MacroServer for this macro

        :return: the MacroServer
        :rtype: :class:`~sardana.macroserver.macroserver.MacroServer`"""
        return self.door.macro_server

    macro_server = property(getMacroServer)

    @mAPI
    def getDoorName(self):
        """**Macro API**.
        Returns the string with the name of the Door that invoked this macro.

        :return: the string with the name of the Door that invoked this macro.
        :rtype: :obj:`str`"""
        return self.door.name

    @mAPI
    def getCommand(self):
        """**Macro API**.
        Returns the string used to execute the macro.
        Ex.: 'ascan M1 0 1000 100 0.8'

        :return: the macro command.
        :rtype: :obj:`str`"""
        return '%s %s' % (self.getName(), ' '.join([str(p) for p in self._in_pars]))

    @mAPI
    def getDateString(self, time_format='%a %b %d %H:%M:%S %Y'):
        """**Macro API**.
        Helper method. Returns the current date in a string.

        :param time_format: the format in which the date should be
                            returned (optional, default value is
                            '%a %b %d %H:%M:%S %Y'
        :type time_format: :obj:`str`
        :return: the current date
        :rtype: :obj:`str`"""
        return time.strftime(time_format)

    @mAPI
    def outputDate(self, time_format='%a %b %d %H:%M:%S %Y'):
        """**Macro API**.
        Helper method. Outputs the current date into the output buffer

        :param time_format: (str) the format in which the date should be
                            returned (optional, default value is
                            '%a %b %d %H:%M:%S %Y'
        :type time_format: :obj:`str`"""
        self.output(self.getDateString(time_format=time_format))

    @mAPI
    def sendRecordData(self, data, codec=None):
        """**Macro API**.
        Sends the given data to the RecordData attribute of the Door

        :param data: (sequence) the data to be sent"""
        self.executor.sendRecordData(data, codec=codec)

    @mAPI
    def plot(self, *args, **kwargs):
        """**Macro API**.
        Sends the plot command to the client using the 'RecordData' DevEncoded
        attribute. The data is encoded using the pickle -> BZ2 codec.

        :param args: the plotting args
        :param kwargs: the plotting keyword args"""
        self.pyplot.plot(*args, **kwargs)
#        data = dict(args=args, kwargs=kwargs)
#        self.sendRecordData(data, codec='bz2_pickle_plot')
            
    @property
    @mAPI
    def pylab(self):
        try:
            pylab = self._pylab
        except AttributeError:
            self._pylab = pylab = self.door.pylab
        return pylab

    @property
    @mAPI
    def pyplot(self):
        try:
            pyplot = self._pyplot
        except AttributeError:
            self._pyplot = pyplot = self.door.pyplot
        return pyplot
    
    @mAPI
    def getData(self):
        """**Macro API**.
        Returns the data produced by the macro. 

        :raises: Exception if no data has been set before on this macro

        :return: the data produced by the macro
        :rtype: object"""
        if not hasattr(self, "_data"):
            raise Exception("Macro '%s' does not produce any data" % self.getName())
        return self._data
    
    @mAPI
    def setData(self, data):
        """**Macro API**. Sets the data for this macro
        
        :param object data: new data to be associated with this macro"""
        self._data = data

    data = property(getData, setData, doc="macro data")

    @mAPI
    def print(self,  *args, **kwargs):
        """**Macro API**.
        Prints a message. Accepted *args* and
        *kwargs* are the same as :func:`print`. Example::

            self.print("this is a print for macro", self.getName())
            
        .. note::
            you will need python >= 3.0. If you have python 2.x then you must
            include at the top of your file the statement::
            
                from __future__ import print_function
        """
        fd = kwargs.get('file', sys.stdout)
        if fd in (sys.stdout, sys.stderr):
            out = StringIO.StringIO()
            kwargs['file'] = out
            end = kwargs.get('end', '\n')
            if end == '\n':
                kwargs['end'] = ''
            ret = print(*args, **kwargs)
            self.output(out.getvalue())
        else:
            ret = print(*args, **kwargs)
        return ret

    @mAPI
    def input(self, msg, *args, **kwargs):
        """**Macro API**.
        If args is present, it is written to standard output without a trailing
        newline. The function then reads a line from input, converts it to a
        string (stripping a trailing newline), and returns that.

        Depending on which type of application you are running, some of the 
        keywords may have no effect (ex.: spock ignores decimals when a number
        is asked).

        Recognized kwargs:

            - data_type : [default: Type.String] specific input type. Can also
              specify a sequence of strings with possible values (use
              allow_multiple=True to say multiple values can be selected)
            - key : [default: no default] variable/label to assign to this input
            - unit: [default: no default] units (useful for GUIs) 
            - timeout : [default: None, meaning wait forever for input]
            - default_value : [default: None, meaning no default value]
              When given, it must be compatible with data_type
            - allow_multiple : [default: False] in case data_type is a
              sequence of values, allow multiple selection
            - minimum : [default: None] When given, must be compatible with data_type (useful for GUIs)
            - maximum : [default: None] When given, must be compatible with data_type (useful for GUIs)
            - step : [default: None] When given, must be compatible with data_type (useful for GUIs)
            - decimals : [default: None] When given, must be compatible with data_type (useful for GUIs)
            
        Examples::

            device_name = self.input("Which device name (%s)?", "tab separated")
            
            point_nb = self.input("How many points?", data_type=Type.Integer)
            
            calc_mode = self.input("Which algorithm?", data_type=["Average", "Integral", "Sum"],
                                   default_value="Average", allow_multiple=False)"""
        if not self.interactive:
            self.warning("Non interactive macro '%s' is asking for input "
                         "(please set this macro interactive to True)",
                         self.getName())
        if self._interactive_mode:
            kwargs['data_type'] = kwargs.get('data_type', Type.String)
            kwargs['allow_multiple'] = kwargs.get('allow_multiple', False)
            kwargs['macro_id'] = self.getID()
            kwargs['macro_name'] = self.getName()
            kwargs['macro'] = self
            return self.getDoorObj().input(msg, *args, **kwargs)
        else:
            if 'default_value' not in kwargs:
                if 'key' not in kwargs:
                    self.warning("%s running in non attended mode was asked "
                                 "for input without default value or key. "
                                 "Returning None")
                    return None
                else:
                    return self.getEnv(kwargs['key'])
            return kwargs['default_value']
    
    @mAPI
    def output(self, msg, *args, **kwargs):
        """**Macro API**.
        Record a log message in this object's output. Accepted *args* and
        *kwargs* are the same as :meth:`logging.Logger.log`.
        Example::

            self.output("this is a print for macro %s", self.getName())

        :param msg: the message to be recorded
        :type msg: :obj:`str`
        :param args: list of arguments
        :param kwargs: list of keyword arguments"""
        return Logger.output(self, msg, *args, **kwargs)

    @mAPI
    def log(self, level, msg, *args, **kwargs):
        """**Macro API**.
        Record a log message in this object's logger. Accepted *args* and
        *kwargs* are the same as :meth:`logging.Logger.log`.
        Example::

            self.debug(logging.INFO, "this is a info log message for macro %s", self.getName())

        :param level: the record level
        :type level: :obj:`int`
        :param msg: the message to be recorded
        :type msg: :obj:`str`
        :param args: list of arguments
        :param kwargs: list of keyword arguments"""
        return Logger.log(self, level, msg, *args, **kwargs)

    @mAPI
    def debug(self, msg, *args, **kwargs):
        """**Macro API**.
        Record a debug message in this object's logger. Accepted *args* and
        *kwargs* are the same as :meth:`logging.Logger.debug`.
        Example::

            self.debug("this is a log message for macro %s", self.getName())

        :param msg: the message to be recorded
        :type msg: :obj:`str`
        :param args: list of arguments
        :param kw: list of keyword arguments"""
        return Logger.debug(self, msg, *args, **kwargs)

    @mAPI
    def info(self, msg, *args, **kwargs):
        """**Macro API**.
        Record an info message in this object's logger. Accepted *args* and
        *kwargs* are the same as :meth:`logging.Logger.info`.
        Example::

            self.info("this is a log message for macro %s", self.getName())

        :param msg: the message to be recorded
        :type msg: :obj:`str`
        :param args: list of arguments
        :param kwargs: list of keyword arguments"""
        return Logger.info(self, msg, *args, **kwargs)

    @mAPI
    def warning(self, msg, *args, **kwargs):
        """**Macro API**.
        Record a warning message in this object's logger. Accepted *args* and
        *kwargs* are the same as :meth:`logging.Logger.warning`.
        Example::

            self.warning("this is a log message for macro %s", self.getName())

        :param msg: the message to be recorded
        :type msg: :obj:`str`
        :param args: list of arguments
        :param kwargs: list of keyword arguments"""
        return Logger.warning(self, msg, *args, **kwargs)

    @mAPI
    def error(self, msg, *args, **kwargs):
        """**Macro API**.
        Record an error message in this object's logger. Accepted *args* and
        *kwargs* are the same as :meth:`logging.Logger.error`.
        Example::

            self.error("this is a log message for macro %s", self.getName())

        :param msg: the message to be recorded
        :type msg: :obj:`str`
        :param args: list of arguments
        :param kwargs: list of keyword arguments
        """
        return Logger.error(self, msg, *args, **kwargs)

    @mAPI
    def critical(self, msg, *args, **kwargs):
        """**Macro API**.
        Record a critical message in this object's logger. Accepted *args* and
        *kwargs* are the same as :meth:`logging.Logger.critical`.
        Example::

            self.critical("this is a log message for macro %s", self.getName())

        :param msg: the message to be recorded
        :type msg: :obj:`str`
        :param args: list of arguments
        :param kwargs: list of keyword arguments"""
        return Logger.critical(self, msg, *args, **kwargs)

    @mAPI
    def trace(self, msg, *args, **kwargs):
        """**Macro API**. Record a trace message in this object's logger.

        :param msg: (str) the message to be recorded
        :param args: list of arguments
        :param kw: list of keyword arguments"""
        return Logger.trace(self, msg, *args, **kwargs)

    @mAPI
    def traceback(self, *args, **kwargs):
        """**Macro API**.
        Logs the traceback with level TRACE on the macro logger."""
        return Logger.traceback(self, *args, **kwargs)

    @mAPI
    def stack(self, *args, **kwargs):
        """**Macro API**.
        Logs the stack with level TRACE on the macro logger."""
        return Logger.stack(self, *args, **kwargs)

    @mAPI
    def report(self, msg, *args, **kwargs):
        """**Macro API**.
        Record a log message in the sardana report (if enabled) with default
        level **INFO**. The msg is the message format string, and the args are
        the arguments which are merged into msg using the string formatting
        operator. (Note that this means that you can use keywords in the
        format string, together with a single dictionary argument.)

        *kwargs* are the same as :meth:`logging.Logger.debug` plus an optional
        level kwargs which has default value **INFO**

        Example::

            self.report("this is an official report of macro %s", self.getName())

        :param msg: the message to be recorded
        :type msg: :obj:`str`
        :param args: list of arguments
        :param kwargs: list of keyword arguments"""
        return self.getDoorObj().report(msg, *args, **kwargs)

    @mAPI
    def flushOutput(self):
        """**Macro API**.
        Flushes the output buffer."""
        return Logger.flushOutput(self)

    @mAPI
    def getMacroThread(self):
        """**Macro API**.
        Returns the python thread where this macro is running

        :return: the python thread where this macro is running
        :rtype: threading.Thread"""
        return self._macro_thread

    @mAPI
    def getMacroThreadID(self):
        """**Macro API**.
        Returns the python thread id where this macro is running

        :return: the python thread id where this macro is running
        :rtype: :obj:`int`"""
        return self.getMacroThread().ident

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Hook helper API
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    @mAPI
    def createExecMacroHook(self, par_str_sequence, parent_macro=None):
        """**Macro API**.
        Creates a hook that executes the macro given as a sequence of strings
        where the first string is macro name and the following strings the
        macro parameters

        :param par_str_sequence: the macro parameters
        :param parent_macro:
            the parent macro object. If None is given (default) then the
            parent macro is this macro

        :return: a ExecMacroHook object (which is a callable object)"""
        return ExecMacroHook(parent_macro or self, par_str_sequence)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Handle child macro execution
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    @mAPI
    def createMacro(self, *pars):
        """**Macro API**. Create a new macro and prepare it for execution
        Several different parameter formats are supported::

            # several parameters:
            self.createMacro('ascan', 'th', '0', '100', '10', '1.0')
            self.createMacro('ascan', 'th', 0, 100, 10, 1.0)
            th = self.getObj('th')
            self.createMacro('ascan', th, 0, 100, 10, 1.0)

            # a sequence of parameters:
            self.createMacro(['ascan', 'th', '0', '100', '10', '1.0')
            self.createMacro(('ascan', 'th', 0, 100, 10, 1.0))
            th = self.getObj('th')
            self.createMacro(['ascan', th, 0, 100, 10, 1.0])

            # a space separated string of parameters:
            self.createMacro('ascan th 0 100 10 1.0')

        :param pars: the command parameters as explained above
        :return:
            a sequence of two elements: the macro object and the result of
            preparing the macro
        :rtype: :obj:`tuple`\<:class:`~sardana.macroserver.macro.Macro`\, seq<obj>>
        """
        return self.prepareMacro(*pars)

    @mAPI
    def prepareMacroObj(self, macro_name_or_klass, *args, **kwargs):
        """**Macro API**. Prepare a new macro for execution

        :param macro_name_or_klass name: name of the macro to be prepared or
                                         the macro class itself
        :param pars: list of parameter objects
        :param init_opts: keyword parameters for the macro constructor
        :param prepare_opts: keyword parameters for the macro prepare

        :return:
            a sequence of two elements: the macro object and the result of
            preparing the macro"""
        # sync our log before calling the child macro prepare in order to avoid
        # mixed outputs between this macro and the child macro
        self.syncLog()
        init_opts = { 'parent_macro' : self }
        return self.executor.prepareMacroObj(macro_name_or_klass, args,
                                             init_opts, kwargs)

    @mAPI
    def prepareMacro(self, *args, **kwargs):
        """**Macro API**. Prepare a new macro for execution
        Several different parameter formats are supported::

            # several parameters:
            executor.prepareMacro('ascan', 'th', '0', '100', '10', '1.0')
            executor.prepareMacro('ascan', 'th', 0, 100, 10, 1.0)
            th = self.getObj('th')
            executor.prepareMacro('ascan', th, 0, 100, 10, 1.0)

            # a sequence of parameters:
            executor.prepareMacro(['ascan', 'th', '0', '100', '10', '1.0')
            executor.prepareMacro(('ascan', 'th', 0, 100, 10, 1.0))
            th = self.getObj('th')
            executor.prepareMacro(['ascan', th, 0, 100, 10, 1.0])

            # a space separated string of parameters:
            executor._prepareMacro('ascan th 0 100 10 1.0')

        :param args: the command parameters as explained above
        :param kwargs: keyword optional parameters for prepare
        :return:
            a sequence of two elements: the macro object and the result of
            preparing the macro"""
        # sync our log before calling the child macro prepare in order to avoid
        # mixed outputs between this macro and the child macro
        self.syncLog()
        init_opts = { 'parent_macro' : self }
        return self.executor.prepareMacro(args, init_opts, kwargs)

    @mAPI
    def runMacro(self, macro_obj):
        """**Macro API**. Runs the macro. This the lower level version of
        :meth:`~sardana.macroserver.macro.Macro.execMacro`. The method only
        returns after the macro is completed or an exception is thrown.
        It should be used instead of execMacro when some operation needs to
        be done between the macro preparation and the macro execution.
        Example::

            macro = self.prepareMacro("mymacro", "myparam")
            self.do_my_stuff_with_macro(macro)
            self.runMacro(macro)

        :param macro_obj: macro object
        :return: the output of the macro"""
        # sync our log before calling the child macro prepare in order to avoid
        # mixed outputs between this macro and the child macro
        self.syncLog()
        return self.executor.runMacro(macro_obj)

    @mAPI
    def execMacroObj(self, name, *args, **kwargs):
        """**Macro API**. Execute a macro in this macro. The method only returns
        after the macro is completed or an exception is thrown. This is a
        higher level version of runMacro method. It is the same as::

            macro = self.prepareMacroObjs(name, *args, **kwargs)
            self.runMacro(macro)
            return macro

        :param name: name of the macro to be prepared
        :type name: str
        :param args: list of parameter objects
        :param kwargs: list of keyword parameters

        :return: a macro object"""
        self.debug("Executing macro: %s" % name)
        macro_obj, prepare_result = self.prepareMacroObj(name, *args, **kwargs)
        self.runMacro(macro_obj)
        return macro_obj

    @mAPI
    def execMacro(self, *args, **kwargs):
        """**Macro API**. Execute a macro in this macro. The method only
        returns after the macro is completed or an exception is thrown. Several
        different parameter formats are supported::

            # several parameters:
            self.execMacro('ascan', 'th', '0', '100', '10', '1.0')
            self.execMacro('ascan', 'th', 0, 100, 10, 1.0)
            th = self.getObj('th')
            self.execMacro('ascan', th, 0, 100, 10, 1.0)

            # a sequence of parameters:
            self.execMacro(['ascan', 'th', '0', '100', '10', '1.0')
            self.execMacro(('ascan', 'th', 0, 100, 10, 1.0))
            th = self.getObj('th')
            self.execMacro(['ascan', th, 0, 100, 10, 1.0])

            # a space separated string of parameters:
            self.execMacro('ascan th 0 100 10 1.0')

        :param pars: the command parameters as explained above

        :return: a macro object"""
        par0 = args[0]
        if len(args) == 1:
            if type(par0) in types.StringTypes :
                args = par0.split()
                
            elif operator.isSequenceType(par0):
                args = par0
        args = map(str, args)

        self.debug("Executing macro: %s" % args[0])
        macro_obj, prepare_result = self.prepareMacro(*args, **kwargs)
        self.runMacro(macro_obj)
        return macro_obj

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # taurus helpers
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    @mAPI
    def getTangoFactory(self):
        """**Macro API**. Helper method that returns the tango factory.

        :return: the tango factory singleton
        :rtype: :class:`~taurus.core.tango.TangoFactory`"""
        import taurus
        return taurus.Factory()

    @mAPI
    def getDevice(self, dev_name):
        """**Macro API**. Helper method that returns the device for the given
        device name

        :return: the taurus device for the given device name
        :rtype: :class:`~taurus.core.TaurusDevice`"""
        import taurus
        return taurus.Device(dev_name)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Handle parameter objects
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    @mAPI
    def setLogBlockStart(self):
        """**Macro API**. Specifies the begining of a block of data. Basically
        it outputs the 'BLOCK' tag"""
        self.output(Macro.BlockStart)

    @mAPI
    def setLogBlockFinish(self):
        """**Macro API**. Specifies the end of a block of data. Basically it
        outputs the '/BLOCK' tag"""
        self.output(Macro.BlockFinish)

    @mAPI
    def outputBlock(self, line):
        """**Macro API**. Sends an line tagged as a block to the output

        :param str line: line to be sent"""
        if isinstance(line, (str, unicode)):
            o = line
        elif operator.isSequenceType(line):
            o = "\n".join(line)
        else:
            o = str(line)
        self.output("%s\n%s\n%s" % (Macro.BlockStart, o, Macro.BlockFinish))

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Handle parameter objects
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    @mAPI
    def getPools(self):
        """**Macro API**. Returns the list of known device pools.

        :return: the list of known device pools
        :rtype: seq<Pool>"""
        return self.door.get_pools()

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Handle parameter objects
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    @mAPI
    def addObj(self, obj, priority=0):
        """**Macro API**. Adds the given object to the list of controlled
        objects of this macro. In practice it means that if a stop is
        executed the stop method of the given object will be called.

        :param obj: the object to be controlled
        :type obj: object
        :param priority:
            wheater or not reserve with priority [default: 0 meaning no
            priority ]
        :type priority: :obj:`int`"""
        self.executor.reserveObj(obj, self, priority=priority)

    @mAPI
    def addObjs(self, obj_list):
        """**Macro API**. Adds the given objects to the list of controlled
        objects of this macro. In practice it means that if a stop is
        executed the stop method of the given object will be called.

        :param obj_list: list of objects to be controlled
        :type obj_list: sequence"""
        for o in obj_list:
            self.addObj(o)

    def returnObj(self, obj):
        """Removes the given objects to the list of controlled objects of this
        macro.

        :param obj: object to be released from the control
        :rtype: object"""
        self.executor.returnObj(obj)

    @mAPI
    def getObj(self, name, type_class=All, subtype=All, pool=All, reserve=True):
        """**Macro API**. Gets the object of the given type belonging to the
        given pool with the given name. The object (if found) will automatically
        become controlled by the macro.

        :raises: MacroWrongParameterType if name is not a string
        :raises: AttributeError if more than one matching object is found

        :param name:
            string representing the name of the object. Can be a regular
            expression
        :type name: :obj:`str`
        :param type_class:
            the type of object [default: All]
        :param subtype:
            a string representing the subtype [default: All]
            Ex.: if type_class is Type.ExpChannel, subtype could be
            'CTExpChannel'
        :param pool:
            the pool to which the object should belong [default: All]
        :param reserve:
            automatically reserve the object for this macro [default: True]

        :return: the object or None if no compatible object is found"""
        if not isinstance(name, (str, unicode)):
            raise self._buildWrongParamExp("getObj", "name", "string",
                                           str(type(name)))

        obj = self.door.get_object(name, type_class=type_class,
                                   subtype=subtype, pool=pool)
        if obj and reserve:
            self.addObj(obj)
        return obj

    @mAPI
    def getObjs(self, names, type_class=All, subtype=All, pool=All, reserve=True):
        """**Macro API**. Gets the objects of the given type belonging to the
           given pool with the given names. The objects (if found) will
           automatically become controlled by the macro.

           :param names: a string or a sequence of strings representing the
                         names of the objects. Each string can be a regular
                         expression
           :param type_class: the type of object (optional, default is All).
                             Example: Type.Motor, Type.ExpChannel
           :param subtype: a string representing the subtype (optional,
                           default is All)
                           Ex.: if type_class is Type.ExpChannel, subtype could
                           be 'CTExpChannel'
           :param pool: the pool to which the object should belong (optional,
                        default is All)
           :param reserve: automatically reserve the object for this macro
                           (optional, default is True)

           :return: a list of objects or empty list if no compatible object is
                    found"""
        obj_list = self.door.get_objects(names, type_class=type_class,
                                         subtype=subtype, pool=pool)
        if reserve:
            self.addObjs(obj_list)
        return obj_list or []

    @mAPI
    def findObjs(self, names, type_class=All, subtype=All, pool=All,
                 reserve=True):
        """**Macro API**. Gets the objects of the given type belonging to the
        given pool with the given names. The objects (if found) will
        automatically become controlled by the macro.

        :param names:
            a string or a sequence of strings representing the names of the
            objects. Each string can be a regular expression
        :param type_class: the type of object (optional, default is All)
        :param subtype:
            a string representing the subtype [default: All]
            Ex.: if type_class is Type.ExpChannel, subtype could be
            'CTExpChannel'
        :param pool:
            the pool to which the object should belong [default: All]
        :param reserve:
            automatically reserve the object for this macro [default: True]

        :return:
            a list of objects or empty list if no compatible object is found"""
        obj_list = self.door.find_objects(names, type_class=type_class,
                                          subtype=subtype, pool=pool)
        if reserve:
            self.addObjs(obj_list)
        return obj_list

    @mAPI
    def getMacroNames(self):
        """**Macro API**. Returns a list of strings containing the names of all
        known macros

           :return: a sequence of macro names
           :rtype: seq<:obj:`str`\>"""
        return self.door.get_macro_names()

    @mAPI
    def getMacros(self, filter=None):
        """**Macro API**. Returns a sequence of
        :class:`~sardana.macroserver.msmetamacro.MacroClass`
        /:class:`~sardana.macroserver.msmetamacro.MacroFunction` objects for all
        known macros that obey the filter expression.

        :param filter:
            a regular expression for the macro name (optional, default is None
            meaning match all macros)
        :return:
            a sequence of :class:`~sardana.macroserver.msmetamacro.MacroClass`
            /:class:`~sardana.macroserver.msmetamacro.MacroFunction`
            objects
        :rtype: seq<:class:`~sardana.macroserver.msmetamacro.MacroClass`
                /:class:`~sardana.macroserver.msmetamacro.MacroFunction`\>"""
        ret = self.door.get_macros(filter=filter).values()
        ret.sort()
        return ret

    @mAPI
    def getMacroLibraries(self, filter=None):
        """**Macro API**. Returns a sequence of
        :class:`~sardana.macroserver.msmetamacro.MacroLibrary` objects for all
        known macros that obey the filter expression.

        :param filter:
            a regular expression for the macro library [default: None meaning
            match all macro libraries)
        :return:
            a sequence of :class:`~sardana.macroserver.msmetamacro.MacroLibrary`
            objects
        :rtype: seq<:class:`~sardana.macroserver.msmetamacro.MacroLibrary`\>"""
        ret = self.door.get_macro_libs(filter=filter).values()
        ret.sort()
        return ret

    @mAPI 
    def getMacroLibrary(self, lib_name):
        """**Macro API**. Returns a
        :class:`~sardana.macroserver.msmetamacro.MacroLibrary` object for the
        given library name.

        :param lib_name:
            library name
        :type lib_name: str
        :return:
            a macro library :class:`~sardana.macroserver.msmetamacro.MacroLibrary`
        :rtype: :class:`~sardana.macroserver.msmetamacro.MacroLibrary`"""
        ret = self.door.get_macro_lib(lib_name)
        return ret

    getMacroLib = getMacroLibrary
    getMacroLibs = getMacroLibraries

    @mAPI
    def getMacroInfo(self, macro_name):
        """**Macro API**. Returns the corresponding
        :class:`~sardana.macroserver.msmetamacro.MacroClass`
        /:class:`~sardana.macroserver.msmetamacro.MacroFunction` object.

        :param macro_name: a string with the desired macro name.
        :type macro_name: :obj:`str`
        :return:
            a :class:`~sardana.macroserver.msmetamacro.MacroClass`
            /:class:`~sardana.macroserver.msmetamacro.MacroFunction` object or
            None if the macro with the given name was not found
        :rtype: :class:`~sardana.macroserver.msmetamacro.MacroClass`
                /:class:`~sardana.macroserver.msmetamacro.MacroFunction`"""
        return self.door.get_macro(macro_name)

    @mAPI
    def getMotion(self, elems, motion_source=None, read_only=False, cache=True):
        """**Macro API**. Returns a new Motion object containing the given
        elements.

        :raises:
            Exception if no elements are defined or the elems is not recognized
            as valid, or an element is not found or an element appears more
            than once

        :param elems: list of moveable object names
        :param motion_source:
            obj or list of objects containing moveable elements. Usually this
            is a Pool object or a list of Pool objects (optional, default is
            None, meaning all known pools will be searched for the given
            moveable items
        :param read_only: not used. Reserved for future use
        :param cache: not used. Reserved for future use

        :return: a Motion object """

        decoupled=False
        try:
            decoupled = self.getEnv("MotionDecoupled")
        except UnknownEnv:
            pass

        motion = self.door.get_motion(elems, motion_source=motion_source,
                                      read_only=read_only, cache=cache,
                                      decoupled=decoupled)
        if motion is not None:
            self.addObj(motion, priority=1)
        return motion

    @mAPI
    def getElementsWithInterface(self, interface):
        return self.door.get_elements_with_interface(interface)

    @mAPI
    def getControllers(self):
        return self.door.get_controllers()

    @mAPI
    def getMoveables(self):
        return self.door.get_moveables()

    @mAPI
    def getMotors(self):
        return self.door.get_motors()

    @mAPI
    def getPseudoMotors(self):
        return self.door.get_pseudo_motors()

    @mAPI
    def getIORegisters(self):
        return self.door.get_io_registers()

    @mAPI
    def getMeasurementGroups(self):
        return self.door.get_measurement_groups()

    @mAPI
    def getExpChannels(self):
        return self.door.get_exp_channels()

    @mAPI
    def getCounterTimers(self):
        return self.door.get_counter_timers()

    @mAPI
    def get0DExpChannels(self):
        return self.door.get_0d_exp_channels()

    @mAPI
    def get1DExpChannels(self):
        return self.door.get_1d_exp_channels()

    @mAPI
    def get2DExpChannels(self):
        return self.door.get_2d_exp_channels()

    @mAPI
    def getPseudoCounters(self):
        return self.door.get_pseudo_counters()

    @mAPI
    def getInstruments(self):
        return self.door.get_instruments()

    @mAPI
    def getElementWithInterface(self, interface, name):
        return self.door.get_element_with_interface(interface, name)

    @mAPI
    def getController(self, name):
        return self.door.get_controller(name)

    @mAPI
    def getMoveable(self, name):
        return self.door.get_moveable(name)

    @mAPI
    def getMotor(self, name):
        return self.door.get_motor(name)

    @mAPI
    def getPseudoMotor(self, name):
        return self.door.get_pseudo_motor(name)

    @mAPI
    def getIORegister(self, name):
        return self.door.get_io_register(name)

    @mAPI
    def getMeasurementGroup(self, name):
        return self.door.get_measurement_group(name)

    @mAPI
    def getExpChannel(self, name):
        return self.door.get_exp_channel(name)

    @mAPI
    def getCounterTimer(self, name):
        return self.door.get_counter_timer(name)

    @mAPI
    def get0DExpChannel(self, name):
        return self.door.get_0d_exp_channel(name)

    @mAPI
    def get1DExpChannel(self, name):
        return self.door.get_1d_exp_channel(name)

    @mAPI
    def get2DExpChannel(self, name):
        return self.door.get_2d_exp_channel(name)

    @mAPI
    def getPseudoCounter(self, name):
        return self.door.get_pseudo_counter(name)

    @mAPI
    def getInstrument(self, name):
        return self.door.get_instrument(name)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Handle macro environment
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    @mAPI
    def getEnv(self, key=None, macro_name=None, door_name=None):
        """**Macro API**. Gets the local environment matching the given
        parameters:

           - door_name and macro_name define the context where to look for
             the environment. If both are None, the global environment is
             used. If door name is None but macro name not, the given macro
             environment is used and so on...
           - If key is None it returns the complete environment, otherwise
             key must be a string containing the environment variable name.

        :raises: UnknownEnv

        :param key:
            environment variable name [default: None, meaning all environment]
        :type key: str
        :param door_name:
            local context for a given door [default: None, meaning no door
            context is used]
        :type door_name: str
        :param macro_name:
            local context for a given macro [default: None, meaning no macro
            context is used]
        :type macro_name: str

        :return: a :obj:`dict` containing the environment
        :rtype: :obj:`dict`"""
        door_name = door_name or self.getDoorName()
        macro_name = macro_name or self._name

        return self.macro_server.get_env(key=key, macro_name=macro_name,
                                         door_name=door_name)

    @mAPI
    def getGlobalEnv(self):
        """**Macro API**. Returns the global environment.

        :return: a :obj:`dict` containing the global environment
        :rtype: :obj:`dict`"""
        return self.macro_server.get_env()

    @mAPI
    def getAllEnv(self):
        """**Macro API**. Returns the enviroment for the macro.

        :return: a :obj:`dict` containing the environment for the macro
        :rtype: :obj:`dict`"""
        return self.getEnv(None)

    @mAPI
    def getAllDoorEnv(self):
        """**Macro API**. Returns the enviroment for the door where the macro
        is running.

        :return: a :obj:`dict` containing the environment
        :rtype: :obj:`dict`"""
        return self.door.get_env()

    @mAPI
    def setEnv(self, key, value):
        """**Macro API**. Sets the environment key to the new value and
        stores it persistently.

        :return: a :obj:`tuple` with the key and value objects stored
        :rtype: :obj:`tuple`\<:obj:`str`\, object>"""
        return self.door.set_env(key, value)

    @mAPI
    def unsetEnv(self, key):
        """**Macro API**. Unsets the given environment variable.

        :param key: the environment variable name
        :type key: :obj:`str`"""
        return self.macro_server.unset_env(key)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Reload API
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    
    @mAPI
    def reloadLibrary(self, lib_name):
        """**Macro API**. Reloads the given library(=module) names

        :raises: ImportError in case the reload process is not
                 successfull

        :param lib_name: library(=module) name
        :type lib_name: :obj:`str`

        :return:
            the reloaded python module object"""
        return self.door.reload_lib(lib_name)    

    @mAPI
    def reloadMacro(self, macro_name):
        """**Macro API**. Reloads the module corresponding to the given macro
        name

        :raises: MacroServerExceptionList in case the macro is unknown or the
                 reload process is not successfull

        :param macro_name: macro name
        :type macro_name: :obj:`str`"""
        return self.door.reload_macro(macro_name)

    @mAPI
    def reloadMacros(self, macro_names):
        """**Macro API**. Reloads the modules corresponding to the given macro
        names.

        :raises: MacroServerExceptionList in case the macro(s) are unknown or
                 the reload process is not successfull

        :param macro_names: a list of macro names
        :type macro_names: sequence<:obj:`str`\>"""
        return self.reload_macros(macro_names)

    @mAPI
    def reloadMacroLibrary(self, lib_name):
        """**Macro API**. Reloads the given library(=module) names

        :raises: MacroServerExceptionList in case the reload process is not
                 successfull

        :param lib_name: library(=module) name
        :type lib_name: :obj:`str`

        :return:
            the :class:`~sardana.macroserver.metamacro.MacroLibrary` for the
            reloaded library
        :rtype: :class:`~sardana.macroserver.metamacro.MacroLibrary`"""
        return self.door.reload_macro_lib(lib_name)

    @mAPI
    def reloadMacroLibraries(self, lib_names):
        """**Macro API**. Reloads the given library(=module) names

        :raises: MacroServerExceptionList in case the reload process is not
                 successfull for at least one lib

        param lib_names: a list of library(=module) names
        :type lib_name: seq<:obj:`str`\>

        :return:
            a sequence of :class:`~sardana.macroserver.metamacro.MacroLibrary`
            objects for the reloaded libraries
        :rtype: seq<:class:`~sardana.macroserver.metamacro.MacroLibrary`\>"""
        return self.door.reload_macro_libs(lib_names)
    
    reloadMacroLib = reloadMacroLibrary
    reloadMacroLibs = reloadMacroLibraries
    
    @mAPI
    def getViewOption(self, name):
        return self._getViewOptions()[name]
    
    @mAPI
    def getViewOptions(self):
        return self._getViewOptions()

    @mAPI
    def setViewOption(self, name, value):
        vo = self._getViewOptions()
        vo[name] = value
        self.setEnv('_ViewOptions', vo)
    
    @mAPI
    def resetViewOption(self, name):
        vo = self._getViewOptions()
        ViewOption.reset_option(vo, name)
        self.setEnv('_ViewOptions', vo)
        return vo.get(name)
    
    #@}

    ## @name Unofficial Macro API
    #    This list contains the set of methods that are <b>NOT</b> part of the
    #  the macro developer knows what he is doing.
    #    Please check before is there is an official API that does the samething
    #  before executing any of these methods.
    #    If you see that your macro needs to execute any of these methods please
    #  consider informing the MacroServer developer so he may expose this in a
    #  safe way.
    #@{
    
    def _getViewOptions(self):
        try:
            vo = self.getEnv('_ViewOptions')
        except UnknownEnv:
            import msoptions
            vo = ViewOption.init_options(dict())
            self.setEnv('_ViewOptions', vo)
        return vo

    def _input(self, msg, *args, **kwargs):
        """**Unofficial Macro API**.
        If args is present, it is written to standard output without a trailing
        newline. The function then reads a line from input, converts it to a
        string (stripping a trailing newline), and returns that.

        Depending on which type of application you are running, some of the 
        keywords may have no effect (ex.: spock ignores decimals when a number
        is asked).

        Recognized kwargs:

            - data_type : [default: Type.String] specific input type. Can also
              specify a sequence of strings with possible values (use
              allow_multiple=True to say multiple values can be selected)
            - key : [default: no default] variable/label to assign to this input
            - unit: [default: no default] units (useful for GUIs) 
            - timeout : [default: None, meaning wait forever for input]
            - default_value : [default: None, meaning no default value]
              When given, it must be compatible with data_type
            - allow_multiple : [default: False] in case data_type is a
              sequence of values, allow multiple selection
            - minimum : [default: None] When given, must be compatible with data_type (useful for GUIs)
            - maximum : [default: None] When given, must be compatible with data_type (useful for GUIs)
            - step : [default: None] When given, must be compatible with data_type (useful for GUIs)
            - decimals : [default: None] When given, must be compatible with data_type (useful for GUIs)
            
        Examples::

            device_name = self.input("Which device name (%s)?", "tab separated")
            
            point_nb = self.input("How many points?", data_type=Type.Integer)
            
            calc_mode = self.input("Which algorithm?", data_type=["Average", "Integral", "Sum"],
                                   default_value="Average", allow_multiple=False)"""
        if not self.interactive:
            self.warning("Non interactive macro '%s' is asking for input "
                         "(please set this macro interactive to True)",
                         self.getName())
        if self._interactive_mode:
            kwargs['data_type'] = kwargs.get('data_type', Type.String)
            kwargs['allow_multiple'] = kwargs.get('allow_multiple', False)
            kwargs['macro_id'] = self.getID()
            kwargs['macro_name'] = self.getName()
            kwargs['macro'] = self
            return self.getDoorObj().input(msg, *args, **kwargs)
        else:
            if 'default_value' not in kwargs:
                if 'key' not in kwargs:
                    self.warning("%s running in non attended mode was asked "
                                 "for input without default value or key. "
                                 "Returning None")
                    return None
                else:
                    return self.getEnv(kwargs['key'])
            return kwargs['default_value']
    
    def _output(self, msg, *args, **kwargs):
        """**Macro API**.
        Record a log message in this object's output. Accepted *args* and
        *kwargs* are the same as :meth:`logging.Logger.log`.
        Example::

            self.output("this is a print for macro %s", self.getName())

        :param msg: the message to be recorded
        :type msg: :obj:`str`
        :param args: list of arguments
        :param kwargs: list of keyword arguments"""
        return Logger.output(self, msg, *args, **kwargs)

    def _log(self, level, msg, *args, **kwargs):
        """**Unofficial Macro API**.
        Record a log message in this object's logger. Accepted *args* and
        *kwargs* are the same as :meth:`logging.Logger.log`.
        Example::

            self.debug(logging.INFO, "this is a info log message for macro %s", self.getName())

        :param level: the record level
        :type level: :obj:`int`
        :param msg: the message to be recorded
        :type msg: :obj:`str`
        :param args: list of arguments
        :param kwargs: list of keyword arguments"""
        return Logger.log(self, level, msg, *args, **kwargs)

    def _debug(self, msg, *args, **kwargs):
        """**Unofficial Macro API**.
        Record a debug message in this object's logger. Accepted *args* and
        *kwargs* are the same as :meth:`logging.Logger.debug`.
        Example::

            self.debug("this is a log message for macro %s", self.getName())

        :param msg: the message to be recorded
        :type msg: :obj:`str`
        :param args: list of arguments
        :param kw: list of keyword arguments"""
        return Logger.debug(self, msg, *args, **kwargs)

    def _info(self, msg, *args, **kwargs):
        """**Unofficial Macro API**.
        Record an info message in this object's logger. Accepted *args* and
        *kwargs* are the same as :meth:`logging.Logger.info`.
        Example::

            self.info("this is a log message for macro %s", self.getName())

        :param msg: the message to be recorded
        :type msg: :obj:`str`
        :param args: list of arguments
        :param kwargs: list of keyword arguments"""
        return Logger.info(self, msg, *args, **kwargs)

    @mAPI
    def _warning(self, msg, *args, **kwargs):
        """**Unofficial Macro API**.
        Record a warning message in this object's logger. Accepted *args* and
        *kwargs* are the same as :meth:`logging.Logger.warning`.
        Example::

            self.warning("this is a log message for macro %s", self.getName())

        :param msg: the message to be recorded
        :type msg: :obj:`str`
        :param args: list of arguments
        :param kwargs: list of keyword arguments"""
        return Logger.warning(self, msg, *args, **kwargs)

    def _error(self, msg, *args, **kwargs):
        """**Unofficial Macro API**.
        Record an error message in this object's logger. Accepted *args* and
        *kwargs* are the same as :meth:`logging.Logger.error`.
        Example::

            self.error("this is a log message for macro %s", self.getName())

        :param msg: the message to be recorded
        :type msg: :obj:`str`
        :param args: list of arguments
        :param kwargs: list of keyword arguments
        """
        return Logger.error(self, msg, *args, **kwargs)

    def _critical(self, msg, *args, **kwargs):
        """**Unofficial Macro API**.
        Record a critical message in this object's logger. Accepted *args* and
        *kwargs* are the same as :meth:`logging.Logger.critical`.
        Example::

            self.critical("this is a log message for macro %s", self.getName())

        :param msg: the message to be recorded
        :type msg: :obj:`str`
        :param args: list of arguments
        :param kwargs: list of keyword arguments"""
        return Logger.critical(self, msg, *args, **kwargs)

    def _trace(self, msg, *args, **kwargs):
        """**Unofficial Macro API**. Record a trace message in this object's logger.

        :param msg: (str) the message to be recorded
        :param args: list of arguments
        :param kw: list of keyword arguments"""
        return Logger.trace(self, msg, *args, **kwargs)

    def _traceback(self, *args, **kwargs):
        """**Unofficial Macro API**.
        Logs the traceback with level TRACE on the macro logger."""
        return Logger.traceback(self, *args, **kwargs)

    def _stack(self, *args, **kwargs):
        """**Unofficial Macro API**.
        Logs the stack with level TRACE on the macro logger."""
        return Logger.stack(self, *args, **kwargs)

    def _report(self, msg, *args, **kwargs):
        """**Unofficial Macro API**.
        Record a log message in the sardana report (if enabled) with default
        level **INFO**. The msg is the message format string, and the args are
        the arguments which are merged into msg using the string formatting
        operator. (Note that this means that you can use keywords in the
        format string, together with a single dictionary argument.)

        *kwargs* are the same as :meth:`logging.Logger.debug` plus an optional
        level kwargs which has default value **INFO**

        Example::

            self.report("this is an official report of macro %s", self.getName())

        :param msg: the message to be recorded
        :type msg: :obj:`str`
        :param args: list of arguments
        :param kwargs: list of keyword arguments"""
        return self.door.report(msg, *args, **kwargs)

    @property
    def executor(self):
        """**Unofficial Macro API**. Alternative to :meth:`getExecutor` that
        does not throw StopException in case of a Stop. This should be
        called only internally"""
        return self._executor

    @property
    def door(self):
        """**Unofficial Macro API**. Alternative to :meth:`getDoorObj` that
        does not throw StopException in case of a Stop. This should be
        called only internally"""
        return self.executor.getDoor()

    @property
    def parent_macro(self):
        """**Unofficial Macro API**. Alternative to getParentMacro that does not
        throw StopException in case of a Stop. This should be called only
        internally by the *Executor*"""
        return self._parent_macro

    @property
    def description(self):
        """**Unofficial Macro API**. Alternative to :meth:`getDescription` that
        does not throw StopException in case of a Stop. This should be
        called only internally by the *Executor*"""
        return self._desc

    def isAborted(self):
        """**Unofficial Macro API**."""
        return self._aborted

    def isStopped(self):
        """**Unofficial Macro API**."""
        return self._stopped

    def isPaused(self):
        """**Unofficial Macro API**."""
        return self._pause_event.isPaused()

    @classmethod
    def hasResult(cls):
        """**Unofficial Macro API**. Returns True if the macro should return
        a result or False otherwise

        :return: True if the macro should return a result or False otherwise
        :rtype: bool
        """
        return len(cls.result_def) > 0

    def getResult(self):
        """**Unofficial Macro API**. Returns the macro result object (if any)

        :return: the macro result object or None"""
        return self._out_pars

    def setResult(self, result):
        """**Unofficial Macro API**. Sets the result of this macro

        :param result: (object) the result for this macro"""
        self._out_pars = result

    ## @name Internal methods
    #  This list contains the set of methods that are for INTERNAL macro usage.
    #  Macro developers should never call any of these methods
    #@{

    @staticmethod
    def _buildWrongParamExp(method_name, param_name, expected, found):
        """**Internal method**. """
        s = "Macro.%s called with wrong parameter type in '%s'. " \
            "Expected %s got %s" % (method_name, param_name, expected, found)
        return MacroWrongParameterType(s)

    def _getName(self):
        """**Internal method**. """
        return self._name

    def _getDescription(self):
        """**Internal method**. """
        return self._desc

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Macro execution methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def _getMacroStatus(self):
        """**Internal method**.
        Returns the current macro status. Macro status is a :obj:`dict` where
        keys are the strings:

            * *id* - macro ID (internal usage only)
            * *range* - the full progress range of a macro (usually a
              :obj:`tuple` of two numbers (0, 100))
            * *state* - the current macro state, a string which can have values
              *start*, *step*, *stop* and *abort*
            * *step* - the current step in macro. Should be a value inside the
              allowed macro range

        :return: the macro status
        :rtype: :obj:`dict`"""
        return self._macro_status

    def _shouldRaiseStopException(self):
        return self.isStopped() and not self.isProcessingStop()

    def _reserveObjs(self, args):
        """**Internal method**. Used to reserve a set of objects for this
        macro"""
        for obj in args:
            # isiterable
            if not type(obj) in map(type,([],())):
            #if not operator.isSequenceType(obj) or type(obj) in types.StringTypes:
                obj = (obj,)
            for sub_obj in obj:
                if isinstance(sub_obj, PoolElement):
                    self.addObj(sub_obj)

    def exec_(self):
        """**Internal method**. Execute macro as an iterator"""
        self._macro_thread = threading.current_thread()
        macro_status = self.getMacroStatus()

        # make sure a 0.0 progress is sent
        yield macro_status

        # allow any macro to be paused at the beginning of its execution
        self.pausePoint()
        
        # Run the macro or obtain a generator 
        res = self.run(*self._in_pars)

        # If macro returns a generator then running the macro means go through
        # the generator steps, otherwise the macro has already ran
        if type(res) == types.GeneratorType:
            it = iter(res)
            for i in it:
                if operator.isMappingType(i):
                    new_range = i.get('range')
                    if new_range is not None:
                        macro_status['range'] = new_range
                    new_step = i.get('step')
                    if new_step is not None:
                        macro_status['step'] = new_step
                elif operator.isNumberType(i):
                    macro_status['step'] = i
                macro_status['state'] = 'step'
                yield macro_status
            # make sure a 'stop' progress is sent in case an exception occurs
            macro_status['state'] = 'stop'
        else:
            self._out_pars = res
            macro_status['step'] = 100.0
        macro_status['state'] = 'finish'
        yield macro_status

    def __prepareResult(self,out):
        """**Internal method**. Decodes the given output in order to be able to
        send to the result channel

        :param out: output value

        :return: the output as a sequence of strings
        """
        if out is None:
            out = ()
        if operator.isSequenceType(out) and not type(out) in types.StringTypes:
            out = map(str,out)
        else:
            out = (str(out),)
        return out

    def _stopOnError(self):
        """**Internal method**. The stop procedure. Calls the user 'on_abort'
        protecting it against exceptions"""
        try:
            self.on_stop()
        except Exception:
            Logger.error(self, "Error in on_stop(): %s", traceback.format_exc())
            Logger.debug(self, "Details: ", exc_info=1)

    def _abortOnError(self):
        """**Internal method**. The stop procedure. Calls the user 'on_abort'
        protecting it against exceptions"""
        try:
            self.on_abort()
        except Exception:
            Logger.error(self, "Error in on_abort(): %s", traceback.format_exc())
            Logger.debug(self, "Details: ", exc_info=1)

    def _pausePoint(self, timeout=None):
        """**Internal method**."""
        if self._pause_event.isPaused():
            self.on_pause()
        self._pause_event.wait(timeout)

    def stop(self):
        """**Internal method**. Activates the stop flag on this macro."""
        self._stopped = True

    def abort(self):
        """**Internal method**. Aborts the macro abruptly."""
        # carefull: Inside this method never call a method that has the
        # mAPI decorator
        Logger.debug(self, "Aborting...")
        self._aborted = True
        ret, i = 0, 0
        while ret != 1:
            self.__resumeForAbort()
            th = self._macro_thread
            th_id = ctypes.c_long(th.ident)
            Logger.debug(self, "Sending AbortException to %s", th.name)
            ret = asyncexc(th_id, ctypes.py_object(AbortException))
            i += 1
            if ret == 0:
                # try again
                if i > 2:
                    self.error("Failed to abort after three tries!")
                    break
                time.sleep(0.1)
            if ret > 1:
                # if it returns a number greater than one, you're in trouble,
                # and you should call it again with exc=NULL to revert the effect
                asyncexc(th_id, None)
                Logger.error(self, "Failed to abort (unknown error code %d)" % ret)
                break

    def setProcessingStop(self, yesno):
        """**Internal method**. Activates the processing stop flag on this
        macro"""
        self._processingStop = yesno

    def isProcessingStop(self):
        """**Internal method**. Checks if this macro is processing stop"""
        return self._processingStop

    def pause(self, cb=None):
        """**Internal method**. Pauses the macro execution. To be called by the
        Door running the macro to pause the current macro"""
        self._pause_event.pause(cb=cb)

    def resume(self, cb=None):
        """**Internal method**. Resumes the macro execution. To be called by
        the Door running the macro to resume the current macro"""
        self._pause_event.resume(cb=cb)

    def __resumeForAbort(self):
        """Called internally to resume the macro execution in case of an abort.
        The macro is resumed but instead of allowing the next user instruction
        to proceed it just waits for an ashyncronous AbortException to be
        thrown"""
        self._pause_event.resumeForAbort()

    #@}

    def __getattr__(self, name):
        try:
            self.door.get_macro(name)
        except UnknownMacro:
            raise AttributeError("%r object has no attribute %r" %
                                 (type(self).__name__, name))
        def f(*args, **kwargs):
            self.syncLog()
            opts = dict(parent_macro=self, executor=self.executor)
            kwargs.update(opts)
            eargs = [name]
            eargs.extend(args)
            return self.execMacro(*eargs, **kwargs)

        setattr(self, name, f)
        return f


class iMacro(Macro):

    interactive = True


class MacroFunc(Macro):

    def __init__(self, *args, **kwargs):
        function = kwargs['function']
        self._function = function
        kwargs['as'] = self._function.func_name
        if function.param_def is not None:
            self.param_def = function.param_def
        if function.result_def is not None:
            self.result_def = function.result_def
        if function.env is not None:
            self.env = function.env
        if function.hints is not None:
            self.hints = function.hints
        if function.interactive is not None:
            self.interactive = function.interactive
        Macro.__init__(self, *args, **kwargs)
        
    def run(self, *args):
        return self._function(self, *args)




##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""Examples of macro functions"""

from __future__ import print_function

__all__ = ["mfunc1", "mfunc2", "mfunc3", "mfunc4", "mfunc5"]

__docformat__ = 'restructuredtext'

from sardana.macroserver.macro import Type, Macro, macro


@macro()
def mfunc1(self):
    """First macro function. No parameters whatsoever"""
    self.output("Executing %s", self.getName())
    self.print("Hello",1)
    self.wa()

@macro()
def mfunc2(self, p1):
    """Second macro function. One parameter of unknown type"""
    self.output("parameter: %s", p1)

@macro([ ["moveable", Type.Moveable, None, "motor to watch"] ])
def mfunc3(self, moveable):
    """Third macro function. A proper moveable parameter"""
    self.output("Moveable %s is at %s", moveable.getName(), moveable.getPosition())
    self.ascan(moveable, 0, 10, 10, 0.1)
    self.mfunc1()
    
@macro()
def mfunc4(self, *args):
    """Fourth macro function. A list of parameters of unknown type"""
    self.output("parameters %s", args)
    
@macro()
def mfunc5(self, *args):
    """Fifth macro function. A list of parameters of unknown type"""
    self.output("parameters %s", args)

@macro([ ["moveable", Type.Moveable, None, "moveable to move"],
         ["position", Type.Float, None, "absolute position"] ])
def move(self, moveable, position):
    """This macro moves a motor to the specified position"""
    moveable.move(position)
    self.print("Motor ended at ", moveable.getPosition())



##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/axisex.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import time
import numpy

from sardana import State
from sardana.pool.controller import OneDController, MaxDimSize
from sardana.pool.controller import DefaultValue, Description, FGet, FSet, Type

def gauss(x, mean, ymax, fwhm, yoffset=0):
    return yoffset + ymax*numpy.power(2,-4*((x-mean)/fwhm)**2)

class Channel:
    
    def __init__(self,idx):
        self.idx = idx            # 1 based axisex
        self.value = []
        self.is_counting = False
        self.active = True
        self.amplitude = BaseValue('1.0')


class BaseValue(object):
    
    def __init__(self, value):
        self.raw_value = value
        self.init()
    
    def init(self):
        self.value = float(self.raw_value)
    
    def get(self):
        return self.value

    def get_value_name(self):
        return self.raw_value


class TangoValue(BaseValue):
    
    def init(self):
        import PyTango
        self.attr_proxy = PyTango.AttributeProxy(self.raw_value)

    def get(self):
        return self.attr_proxy.read().value


class DummyOneDController(OneDController):
    "This class is the Tango Sardana OneDController controller for tests"

    gender = "Simulation"
    model  = "Basic"
    organization = "Sardana team"

    MaxDevice = 1024
    
    BufferSize = 1024,

    axis_attributes = {
        'Amplitude' : { 
            Type : str,
            FGet : 'getAmplitude', 
            FSet : 'setAmplitude',        
            Description : 'Amplitude. Maybe a number or a tango attribute(must start with tango://)',
            DefaultValue : '1.0' },
    }
    
    def __init__(self, inst, props, *args, **kwargs):
        OneDController.__init__(self, inst, props, *args, **kwargs)
        self.channels = self.MaxDevice*[None,]
        self.reset()

    def GetAxisAttributes(self, axis):
        # the default max shape for 'value' is (16*1024,). We don't need so much
        # so we set it to BufferSize
        attrs = super(DummyOneDController, self).GetAxisAttributes(axis)
        attrs['Value'][MaxDimSize] = self.BufferSize
        return attrs
        
    def reset(self):
        self.start_time = None
        self.integ_time = None
        self.monitor_count = None
        self.read_channels = {}
        self.counting_channels = {}
        
    def AddDevice(self,axis):
        idx = axis - 1
        self.channels[idx] = channel = Channel(axis)
        channel.value = numpy.zeros(self.BufferSize, dtype=numpy.float64)
        
    def DeleteDevice(self,axis):
        idx = axis - 1
        self.channels[idx] = None

    def PreStateAll(self):
        pass
    
    def PreStateOne(self, axis):
        pass
    
    def StateAll(self):
        pass
    
    def StateOne(self, axis):
        idx = axis - 1
        sta = State.On
        status = "Stopped"
        if axis in self.counting_channels:
            channel = self.channels[idx]
            now = time.time()
            elapsed_time = now - self.start_time
            self._updateChannelState(axis, elapsed_time)
            if channel.is_counting:
                sta = State.Moving
                status = "Acquiring"
        return sta, status
        
    def _updateChannelState(self, axis, elapsed_time):
        channel = self.channels[axis-1]
        if self.integ_time is not None:
            # counting in time
            if elapsed_time >= self.integ_time:
                self._finish(elapsed_time)
        elif self.monitor_count is not None:
            # monitor counts
            v = int(elapsed_time*100*axis)
            if v >= self.monitor_count:
                self._finish(elapsed_time)
    
    def _updateChannelValue(self, axis, elapsed_time):
        channel = self.channels[axis-1]
        t = elapsed_time
        if self.integ_time is not None and not channel.is_counting:
            t = self.integ_time
        x = numpy.linspace(-10, 10, self.BufferSize[0])
        amplitude = axis * t * channel.amplitude.get()
        channel.value = gauss(x, 0, amplitude, 4)
    
    def _finish(self, elapsed_time, axis=None):
        if axis is None:
            for axis, channel in self.counting_channels.items():
                channel.is_counting = False
                self._updateChannelValue(axis, elapsed_time)
        else:
            if axis in self.counting_channels:
                channel = self.counting_channels[axis]
                channel.is_counting = False
                self._updateChannelValue(axis, elapsed_time)
            else:
                channel = self.channels[axis-1]
                channel.is_counting = False
        self.counting_channels = {}
                
    def PreReadAll(self):
        self.read_channels = {}
    
    def PreReadOne(self,axis):
        channel = self.channels[axis-1]
        self.read_channels[axis] = channel

    def ReadAll(self):
        # if in acquisition then calculate the values to return
        if self.counting_channels:
            now = time.time()
            elapsed_time = now - self.start_time
            for axis, channel in self.read_channels.items():
                self._updateChannelState(axis, elapsed_time)
                if channel.is_counting:
                    self._updateChannelValue(axis, elapsed_time)
    
    def ReadOne(self, axis):
        self._log.debug("ReadOne(%s)", axis)
        v = self.read_channels[axis].value
        return v
    
    def PreStartAll(self):
        self.counting_channels = {}
    
    def PreStartOne(self, axis, value):
        idx = axis - 1
        channel = self.channels[idx]
        channel.value = 0.0
        self.counting_channels[axis] = channel
        return True
    
    def StartOne(self, axis, value):
        self.counting_channels[axis].is_counting = True
    
    def StartAll(self):
        self.start_time = time.time()
    
    def LoadOne(self, axis, value):
        idx = axis - 1
        if value > 0:
            self.integ_time = value
            self.monitor_count = None
        else:
            self.integ_time = None
            self.monitor_count = -value
    
    def AbortOne(self, axis):
        now = time.time()
        if axis in self.counting_channels:
            elapsed_time = now - self.start_time
            self._finish(elapsed_time, axis=axis)
    
    def getAmplitude(self, axis):
        idx = axis - 1
        channel = self.channels[idx]
        return channel.amplitude.get_value_name()
    
    def setAmplitude(self, axis, value):
        idx = axis - 1
        channel = self.channels[idx]
        
        klass = BaseValue
        if value.startswith("tango://"):
            klass = TangoValue
        channel.amplitude = klass(value) 
        



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Sardana libray. It defines the base classes
for Sardana values"""

__all__ = ["SardanaValue"]

__docformat__ = 'restructuredtext'

import time


class SardanaValue(object):
    
    def __init__(self, value=None, exc_info=None, timestamp=None,
                 dtype=None, dformat=None):
        self.value = value
        self.error = exc_info is not None
        self.exc_info = exc_info 
        if timestamp is None:
            timestamp = time.time()
        self.timestamp = timestamp
        self.dtype = dtype
        self.dformat = dformat
        
    def __repr__(self):
        v = None
        if self.error:
            v = "<Error>"
        else:
            v = self.value
        return "{0.__class__.__name__}(value={1}, timestamp={0.timestamp})".format(self, v)

    def __str__(self):
        return repr(self)



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Pool library. It defines the base classes
for"""

__all__ = [ "PoolMotorGroup" ]

__docformat__ = 'restructuredtext'

import time
import collections

from sardana import ElementType
from sardana.sardanaattribute import SardanaAttribute

from .poolgroupelement import PoolGroupElement
from .poolmotion import PoolMotion


class Position(SardanaAttribute):
    
    def __init__(self, *args, **kwargs):
        self._w_value_map = None
        super(Position, self).__init__(*args, **kwargs)
        for pos_attr in self.obj.get_physical_position_attribute_iterator():
            pos_attr.add_listener(self.on_change)
                
    def _has_value(self):
        for pos_attr in self.obj.get_physical_position_attribute_iterator():
            if not pos_attr.has_value():
                return False
        return True
    
    def _in_error(self):
        for pos_attr in self.obj.get_physical_position_attribute_iterator():
            if pos_attr.in_error():
                return True
        return False
    
    def get_elements(self):
        return self.obj.get_user_elements()
    
    def get_element_nb(self):
        return len(self.get_user_elements())    
   
    def _get_exc_info(self):
        for position_attr in self.obj.get_physical_position_attribute_iterator():
            if position_attr.error:
                return position_attr.get_exc_info()
    
    def _get_timestamp(self):
        return max( [ pos_attr.timestamp for pos_attr in self.obj.get_physical_position_attribute_iterator() ] )

    def _set_value(self, value, exc_info=None, timestamp=None, propagate=1):
        raise Exception("Cannot set position value for motor group %s" % self.obj.name)
    
    def _get_value(self):
        return [ position.value for position in self.obj.get_physical_position_attribute_iterator() ]

    def _get_write_value(self):
        return [ position.w_value for position in self.obj.get_physical_position_attribute_iterator() ]

    def _set_write_value(self, w_value, timestamp=None, propagate=1):
        assert len(w_value) == self.get_element_nb()
        if isinstance(w_value, collections.Sequence):
            w_value_map = {}
            for v, elem in zip(w_value, self.get_elements()):
                w_value_map[elem] = v
        else:
            w_value_map = w_value
            w_value = []
            for elem in self.get_elements():
                w_value.append(w_value_map[elem])
        self._w_value_map = w_value
        super(Position, self).set_write_value(w_value, timestamp=timestamp,
                                              propagate=propagate) 

    def on_change(self, evt_src, evt_type, evt_value):
        self.fire_read_event(propagate=evt_type.priority)

    def update(self, cache=True, propagate=1):
        if cache:
            for phy_elem_pos in self.obj.get_low_level_physical_position_attribute_iterator():
                if not phy_elem_pos.has_value():
                    cache = False
                    break
        if not cache:
            dial_position_values = self.obj.motion.read_dial_position(serial=True)
            for motion_obj, position_value in dial_position_values.items():
                motion_obj.put_dial_position(position_value, propagate=propagate)


class PoolMotorGroup(PoolGroupElement):

    def __init__(self, **kwargs):
        self._physical_elements = []
        self._in_start_move = False
        kwargs['elem_type'] = ElementType.MotorGroup
        PoolGroupElement.__init__(self, **kwargs)
        on_change = self.on_change
        self._position = Position(self, listeners=on_change)

    def _create_action_cache(self):
        motion_name = "%s.Motion" % self._name
        return PoolMotion(self, motion_name)

    # --------------------------------------------------------------------------
    # Event forwarding
    # --------------------------------------------------------------------------
    
    def on_change(self, evt_src, evt_type, evt_value):
        # forward all events coming from attributes to the listeners
        self.fire_event(evt_type, evt_value)

    def on_element_changed(self, evt_src, evt_type, evt_value):
        name = evt_type.name.lower()
        if name in ('state', 'position'):
            state, status = self._calculate_states()
            if name == 'state':
                propagate_state = evt_type.priority
            else:
                propagate_state = 0
            self.set_state(state, propagate=propagate_state)
            self.set_status(status, propagate=propagate_state)
                
    def add_user_element(self, element, index=None):
        elem_type = element.get_type()
        if elem_type == ElementType.Motor:
            pass
        elif elem_type == ElementType.PseudoMotor:
            #TODO: make this happen
            pass
        else:
            raise Exception("element %s is not a motor" % element.name)

        PoolGroupElement.add_user_element(self, element, index=index)

    # --------------------------------------------------------------------------
    # position
    # --------------------------------------------------------------------------

    def get_position_attribute(self):
        return self._position

    def get_low_level_physical_position_attribute_iterator(self):
        return self.get_physical_elements_attribute_iterator()
        
    def get_physical_position_attribute_iterator(self):
        return self.get_user_elements_attribute_iterator()
    
    def get_physical_positions_attribute_sequence(self):
        return self.get_user_elements_attribute_sequence()
    
    def get_physical_positions_attribute_map(self):
        return self.get_user_elements_attribute_map()

    def get_position(self, cache=True, propagate=1):
        """Returns the user position.

        :param cache:
            if ``True`` (default) return value in cache, otherwise read value
            from hardware
        :type cache:
            bool
        :param propagate:
            0 for not propagating, 1 to propagate, 2 propagate with priority
        :type propagate:
            int
        :return:
            the user position
        :rtype:
            :class:`~sardana.sardanaattribute.SardanaAttribute`"""
        position = self._position
        position.update(cache=cache, propagate=propagate)
        return position        

    def set_position(self, positions):
        """Moves the motor group to the specified user positions

        :param positions:
            the user positions to move to
        :type positions:
            sequence< :class:`~numbers.Number` >"""        
        self.start_move(positions)

    def set_write_position(self, w_position, timestamp=None, propagate=1):
        """Sets a new write value for the user position.

        :param w_position:
            the new write value for user position
        :type w_position:
            sequence< :class:`~numbers.Number` >
        :param propagate:
            0 for not propagating, 1 to propagate, 2 propagate with priority
        :type propagate: int"""
        self._position.set_write_value(w_position, timestamp=timestamp,
                                       propagate=propagate)

    position = property(get_position, set_position, doc="motor group positions")

    # --------------------------------------------------------------------------
    # default acquisition channel
    # --------------------------------------------------------------------------

    def get_default_attribute(self):
        return self.get_position_attribute()
    
    # --------------------------------------------------------------------------
    # motion
    # --------------------------------------------------------------------------

    def get_motion(self):
        return self.get_action_cache()

    motion = property(get_motion, doc="motion object")

    # --------------------------------------------------------------------------
    # motion calculation
    # --------------------------------------------------------------------------

    def calculate_motion(self, new_positions, items=None):
        user_elements = self.get_user_elements()
        if items is None:
            items = {}
        calculated = {}
        for new_position, element in zip(new_positions, user_elements):
            calculated[element] = new_position

        for new_position, element in zip(new_positions, user_elements):
            element.calculate_motion(new_position, items=items,
                                     calculated=calculated)
        return items

    def start_move(self, new_position):
        self._in_start_move = True
        try:
            return self._start_move(new_position)
        finally:
            self._in_start_move = False
            
    def _start_move(self, new_positions):
        self._aborted = False
        items = self.calculate_motion(new_positions)
        timestamp = time.time()
        for item, position_info in items.items():
            item.set_write_position(position_info[0], timestamp=timestamp,
                                    propagate=0)
        if not self._simulation_mode:
            self.motion.run(items=items)



##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This file contains the code for an hypothetical Springfield motor controller
used in documentation"""

import springfieldlib

from sardana import State
from sardana.pool.controller import MotorController

class SpringfieldBaseMotorController(MotorController):
    """The most basic controller intended from demonstration purposes only.
    This is the absolute minimum you have to implement to set a proper motor
    controller able to get a motor position, get a motor state and move a
    motor.
    
    This example is so basic that it is not even directly described in the
    documentation"""

    MaxDevice = 128
    
    def __init__(self, inst, props, *args, **kwargs):
        """Constructor"""
        super(SpringfieldBaseMotorController, self).__init__(inst, props, *args, **kwargs)
        self.springfield = springfieldlib.SpringfieldMotorHW()
        
    def ReadOne(self, axis):
        """Get the specified motor position"""
        return self.springfield.getPosition(axis)
        
    def StateOne(self, axis):
        """Get the specified motor state"""
        springfield = self.springfield
        state = springfield.getState(axis)
        if state == 1:
            return State.On, "Motor is stopped"
        elif state == 2:
            return State.Moving, "Motor is moving"
        elif state == 3:
            return State.Fault, "Motor has an error"
    
    def StartOne(self, axis, position):
        """Move the specified motor to the specified position"""
        self.springfield.move(axis, position)              

    def StopOne(self, axis):
        """Stop the specified motor"""
        self.springfield.stop(axis)        


from sardana import DataAccess
from sardana.pool.controller import Type, Description, DefaultValue, Access, FGet, FSet
    
class SpringfieldMotorController(MotorController):

    axis_attributes = { 
        "CloseLoop" : {
                Type         : bool,
                Description  : "(de)activates the motor close loop algorithm",
                DefaultValue : False,
            },
    }
    
    def getCloseLoop(self, axis):
        return self.springfield.isCloseLoopActive(axis)
    
    def setCloseLoop(self, axis, value):
        self.springfield.setCloseLoop(axis, value)
            
    def __init__(self, inst, props, *args, **kwargs):
        super(SpringfieldMotorController, self).__init__(inst, props, *args, **kwargs)
        
        # initialize hardware communication
        self.springfield = springfieldlib.SpringfieldMotorHW()
        
        # do some initialization
        self._motors = {}

    def AddDevice(self, axis):
        self._motors[axis] = True 

    def DeleteDevice(self, axis):
        del self._motors[axis]        
        
    StateMap = {
        1 : State.On,
        2 : State.Moving,
        3 : State.Fault,
    }
    
    def StateOne(self, axis):
        springfield = self.springfield
        state = self.StateMap[ springfield.getState(axis) ]
        status = springfield.getStatus(axis)
        
        limit_switches = MotorController.NoLimitSwitch
        hw_limit_switches = springfield.getLimits(axis)
        if hw_limit_switches[0]:
            limit_switches |= MotorController.HomeLimitSwitch
        if hw_limit_switches[1]:
            limit_switches |= MotorController.UpperLimitSwitch
        if hw_limit_switches[2]:
            limit_switches |= MotorController.LowerLimitSwitch
        return state, status, limit_switches

    def ReadOne(self, axis):
        position = self.springfield.getPosition(axis)
        return position
        
    def StartOne(self, axis, position):
        self.springfield.move(axis, position)        

    def StopOne(self, axis):
        self.springfield.stop(axis)  

    def AbortOne(self, axis):
        self.springfield.abort(axis)
        
    def DefinePosition(self, axis, position):
        self.springfield.setCurrentPosition(axis, position)



##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains macros that demonstrate the usage of hooks"""

__all__ = ["captain_hook", "captain_hook2", "loop", "hooked_scan",
           "hooked_dummyscan"]

__docformat__ = 'restructuredtext'

from sardana.macroserver.macro import *
        
class loop(Macro, Hookable):
    """A macro that executes a for loop. It accepts hooks.
    This macro is part of the examples package. It was written for 
    demonstration purposes"""
       
    hints = { 'allowsHooks':('pre-move', 'post-move', 'pre-acq', 'post-acq') }
    
    param_def = [['start', Type.Integer, None, 'start point'],
                 ['stop', Type.Integer, None, 'end point'],
                 ['step', Type.Integer, 1, 'step']]

    def hook1(self):
        self.output("En hook 1")
        
    def run(self, start, stop, step):
        self.info("Starting loop")
        self.hooks = [ (self.hook1, ["pre-acq"])]
        for i in xrange(start, stop, step):
            self.output("At step %d" % i)
            self.flushOutput()
            
            for hook,hints in self.hooks:
                self.info("running hook with hints="+repr(hints))
                hook()
        self.info("Finished loop")

class captain_hook(Macro):
    """A macro that executes a loop macro. A hook was attached so that in each 
    step of the loop this hook is executed.
    This macro is part of the examples package. It was written for 
    demonstration purposes"""

    param_def = [['start', Type.Integer, None, 'start point'],
                 ['stop', Type.Integer, None, 'end point'],
                 ['step', Type.Integer, 1, 'step']]
    
    def hook(self):
        self.info("\thook execution")
    
    def run(self, start, stop, step):
        loop_macro = self.createMacro("loop",start,stop,step)
        loop_macro.hooks = [ self.hook ]
        self.runMacro(loop_macro)

class captain_hook2(Macro):
    """A macro that executes a loop macro. A hook was attached so that in each 
    step of the loop this hook is executed.
    This macro is part of the examples package. It was written for 
    demonstration purposes"""

    param_def = [['start', Type.Integer, None, 'start point'],
                 ['stop', Type.Integer, None, 'end point'],
                 ['step', Type.Integer, 1, 'step']]
    
    def hook(self):
        self.execMacroStr(["lsm"])
    
    def run(self, start, stop, step):
        loop_macro = self.createMacro("loop",start,stop,step)
        #h = self.createExecMacroHook(["lsm"])
        loop_macro.hooks = [ self.hook ]  #it gives the "pre-acq" hint to the hook
        self.runMacro(loop_macro)

class hooked_scan(Macro):
    """An example on how to attach hooks to the various hook points of a scan.
    This macro is part of the examples package. It was written for 
    demonstration purposes"""
    
    param_def = [
       ['motor',      Type.Moveable,None, 'Motor to move'],
       ['start_pos',  Type.Float,   None, 'Scan start position'],
       ['final_pos',  Type.Float,   None, 'Scan final position'],
       ['nr_interv',  Type.Integer, None, 'Number of scan intervals'],
       ['integ_time', Type.Float,   None, 'Integration time']
    ]
    def hook1(self):
        self.info("\thook1 execution")
    
    def hook2(self):
        self.info("\thook2 execution")
    
    def hook3(self):
        self.info("\thook3 execution")
    
    def hook4(self):
        self.info("\thook4 execution")
    
    def hook5(self):
        self.info("\thook5 execution")
    
    def hook6(self):
        self.info("\thook6 execution")
    
    def run(self, motor, start_pos, final_pos, nr_interv, integ_time):
        ascan, pars = self.createMacro("ascan",motor, start_pos, final_pos, nr_interv, integ_time)
        ascan.hooks = [ (self.hook1, ["pre-acq"]), 
                       (self.hook2, ["pre-acq","post-acq","pre-move", "post-move","aaaa"]), 
                       self.hook3,
                       (self.hook4, ["pre-scan"]),
                       (self.hook5, ["pre-scan", "post-scan"]),
                       (self.hook6, ["post-step"])]
        self.runMacro(ascan)


class hooked_dummyscan(Macro):
    """An example on how to attach hooks to the various hook points of a scan.
    This macro is part of the examples package. It was written for 
    demonstration purposes"""
    
    param_def = [
       ['start_pos',  Type.Float,   None, 'Scan start position'],
       ['final_pos',  Type.Float,   None, 'Scan final position'],
       ['nr_interv',  Type.Integer, None, 'Number of scan intervals'],
       ['integ_time', Type.Float,   None, 'Integration time']
    ]
    def hook1(self):
        self.info("\thook1 execution")
    
    def hook2(self):
        self.info("\thook2 execution")
    
    def hook3(self):
        self.info("\thook3 execution")
    
    def run(self, start_pos, final_pos, nr_interv, integ_time):
        dummyscan,pars = self.createMacro("dummyscan",start_pos, final_pos, nr_interv, integ_time)
        dummyscan.hooks = [ (self.hook1, ["pre-scan"]), (self.hook2, ["pre-acq","post-acq","pre-move", "post-move","aaaa"]),(self.hook3, ["post-scan"])]
        self.runMacro(dummyscan)




##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This is the standard macro module"""

__all__ = ["ct", "mstate", "mv", "mvr", "pwa", "pwm", "set_lim", "set_lm",
           "set_pos", "settimer", "uct", "umv", "umvr", "wa", "wm"]

__docformat__ = 'restructuredtext'

import datetime
from taurus.console.table import Table

from PyTango import DevState
from sardana.macroserver.macro import Macro, macro, Type, ParamRepeat, ViewOption

################################################################################
#
# Motion related macros
#
################################################################################

class _wm(Macro):
    """Show motor positions"""

    param_def = [
        ['motor_list',
         ParamRepeat(['motor', Type.Moveable, None, 'Motor to move']),
         None, 'List of motor to show'],
    ]

    def prepare(self, *motor_list, **opts):
        self.table_opts = {}
    
    def run(self, *motor_list):
        show_dial = self.getViewOption(ViewOption.ShowDial)
        motor_width = 9
        motor_names = []
        motor_pos   = []
        motor_list = list(motor_list)
        motor_list.sort()
        for motor in motor_list:
            name = motor.getName()
            motor_names.append([name])
            pos = motor.getPosition(force=True)
            if pos is None:
                pos = float('NAN')
            
            if show_dial:
                dial_pos = motor.getDialPosition(force=True)
                if dial_pos is None:
                    dial_pos = float('NAN')
                motor_pos.append((pos,dial_pos))
            else:
                motor_pos.append((pos,))

            motor_width = max(motor_width,len(name))

        fmt = '%c*.%df' % ('%',motor_width - 5)

        table = Table(motor_pos, elem_fmt=[fmt],
                      col_head_str=motor_names, col_head_width=motor_width,
                      **self.table_opts)
        for line in table.genOutput():
            self.output(line)

class _wum(Macro):
    """Show user motor positions"""

    param_def = [
        ['motor_list',
         ParamRepeat(['motor', Type.Moveable, None, 'Motor to move']),
         None, 'List of motor to show'],
    ]

    def prepare(self, *motor_list, **opts):
        self.table_opts = {}
    
    def run(self, *motor_list):
        show_dial = self.getViewOption(ViewOption.ShowDial)
        motor_width = 9
        motor_names = []
        motor_pos   = []
        motor_list = list(motor_list)
        motor_list.sort()
        for motor in motor_list:
            name = motor.getName()
            motor_names.append([name])
            pos = motor.getPosition(force=True)
            if pos is None:
                pos = float('NAN')
            motor_pos.append((pos,))
            motor_width = max(motor_width,len(name))

        fmt = '%c*.%df' % ('%',motor_width - 5)

        table = Table(motor_pos, elem_fmt=[fmt],
                      col_head_str=motor_names, col_head_width=motor_width,
                      **self.table_opts)
        for line in table.genOutput():
            self.output(line)
            
class wu(Macro):
    """Show all user motor positions"""

    def prepare(self, **opts):
        self.all_motors = self.findObjs('.*', type_class=Type.Moveable)
        self.table_opts = {}
    
    def run(self):
        nr_motors = len(self.all_motors)
        if nr_motors == 0:
            self.output('No motor defined')
            return
        
        self.output('Current positions (user) on %s'%datetime.datetime.now().isoformat(' '))
        self.output('')
        
        self.execMacro('_wum',*self.all_motors, **self.table_opts)

class wa(Macro):
    """Show all motor positions"""

    def prepare(self, **opts):
        self.all_motors = self.findObjs('.*', type_class=Type.Moveable)
        self.table_opts = {}
    
    def run(self):
        nr_motors = len(self.all_motors)
        if nr_motors == 0:
            self.output('No motor defined')
            return
        
        show_dial = self.getViewOption(ViewOption.ShowDial)
        if show_dial:
            self.output('Current positions (user, dial) on %s'%datetime.datetime.now().isoformat(' '))
        else:
            self.output('Current positions (user) on %s'%datetime.datetime.now().isoformat(' '))
        self.output('')
        
        self.execMacro('_wm',*self.all_motors, **self.table_opts)


class pwa(Macro):
    """Show all motor positions in a pretty table"""

    def run(self):
        self.execMacro('wa', **Table.PrettyOpts)

class set_lim(Macro):
    """Sets the software limits on the specified motor hello"""
    param_def = [
        ['motor', Type.Moveable, None, 'Motor name'],
        ['low',   Type.Float, None, 'lower limit'],
        ['high',   Type.Float, None, 'upper limit']
    ]
    
    def run(self, motor, low, high):
        name = motor.getName()
        self.debug("Setting user limits for %s" % name)
        motor.getPositionObj().setLimits(low,high)
        self.output("%s limits set to %.4f %.4f (user units)" % (name, low, high))

class set_lm(Macro):
    """Sets the dial limits on the specified motor"""
    param_def = [
        ['motor', Type.Motor, None, 'Motor name'],
        ['low',   Type.Float, None, 'lower limit'],
        ['high',   Type.Float, None, 'upper limit']
    ]
    
    def run(self, motor, low, high):
        name = motor.getName()
        self.debug("Setting dial limits for %s" % name)
        motor.getDialPositionObj().setLimits(low,high)
        self.output("%s limits set to %.4f %.4f (dial units)" % (name, low, high))
        
class set_pos(Macro):
    """Sets the position of the motor to the specified value"""
    
    param_def = [
        ['motor', Type.Motor, None, 'Motor name'],
        ['pos',   Type.Float, None, 'Position to move to']
    ]
    
    def run(self, motor, pos):
        name = motor.getName()
        old_pos = motor.getPosition(force=True)
        motor.definePosition(pos)
        self.output("%s reset from %.4f to %.4f" % (name, old_pos, pos))

class set_user_pos(Macro):
    """Sets the USER position of the motor to the specified value (by changing OFFSET and keeping DIAL)"""
    
    param_def = [
        ['motor', Type.Motor, None, 'Motor name'],
        ['pos',   Type.Float, None, 'Position to move to']
    ]
    
    def run(self, motor, pos):
        name = motor.getName()
        old_pos = motor.getPosition(force=True)
        offset_attr = motor.getAttribute('Offset')
        old_offset = offset_attr.read().value
        new_offset = pos - (old_pos - old_offset)
        offset_attr.write(new_offset)
        self.output("%s reset from %.4f (offset %.4f) to %.4f (offset %.4f)" % (name, old_pos, old_offset, pos, new_offset))

class wm(Macro):
    """Show the position of the specified motors."""

    param_def = [
        ['motor_list',
         ParamRepeat(['motor', Type.Moveable, None, 'Motor to see where it is']),
         None, 'List of motor to show'],
    ]

    def prepare(self, *motor_list, **opts):
        self.table_opts = {}
        
    def run(self, *motor_list):
        motor_width = 10
        motor_names = []
        motor_pos   = []

        show_dial = self.getViewOption(ViewOption.ShowDial)
        
        for motor in motor_list:
            name = motor.getName()
            motor_names.append([name])
            posObj = motor.getPositionObj()
            upos = map(str, [posObj.getMaxValue(), motor.getPosition(force=True), posObj.getMinValue()])
            pos_data = [''] + upos
            if show_dial:
                dPosObj = motor.getDialPositionObj()
                dpos = map(str, [dPosObj.getMaxValue(), motor.getDialPosition(force=True), dPosObj.getMinValue()])
                pos_data += [''] + dpos
            
            motor_pos.append(pos_data)

        elem_fmt = (['%*s'] + ['%*s'] * 3) * 2
        row_head_str = ['User', ' High', ' Current', ' Low']
        if show_dial:
            row_head_str += ['Dial', ' High', ' Current', ' Low']
        table = Table(motor_pos, elem_fmt=elem_fmt, row_head_str=row_head_str,
                      col_head_str=motor_names, col_head_width=motor_width,
                      **self.table_opts)
        for line in table.genOutput():
            self.output(line)

class wum(Macro):
    """Show the user position of the specified motors."""

    param_def = [
        ['motor_list',
         ParamRepeat(['motor', Type.Moveable, None, 'Motor to see where it is']),
         None, 'List of motor to show'],
    ]

    def prepare(self, *motor_list, **opts):
        self.table_opts = {}
        
    def run(self, *motor_list):
        motor_width = 10
        motor_names = []
        motor_pos   = []
        
        for motor in motor_list:
            name = motor.getName()
            motor_names.append([name])
            posObj = motor.getPositionObj()
            upos = map(str, [posObj.getMaxValue(), motor.getPosition(force=True), posObj.getMinValue()])
            pos_data = [''] + upos
            
            motor_pos.append(pos_data)

        elem_fmt = (['%*s'] + ['%*s'] * 3) * 2
        row_head_str = ['User', ' High', ' Current', ' Low',]
        table = Table(motor_pos, elem_fmt=elem_fmt, row_head_str=row_head_str,
                      col_head_str=motor_names, col_head_width=motor_width,
                      **self.table_opts)
        for line in table.genOutput():
            self.output(line)
            
class pwm(Macro):
    """Show the position of the specified motors in a pretty table"""

    param_def = [
        ['motor_list',
         ParamRepeat(['motor', Type.Moveable, None, 'Motor to move']),
         None, 'List of motor to show'],
    ]

    def run(self, *motor_list):
        self.execMacro('wm', *motor_list, **Table.PrettyOpts)

class mv(Macro):
    """Move motor(s) to the specified position(s)"""

    param_def = [
       ['motor_pos_list',
        ParamRepeat(['motor', Type.Moveable, None, 'Motor to move'],
                    ['pos',   Type.Float, None, 'Position to move to']),
        None, 'List of motor/position pairs'],
    ]

    def run(self, *motor_pos_list):
        motors, positions = [], []
        for motor, pos in motor_pos_list:
            motors.append(motor)
            positions.append(pos)
            self.debug("Starting %s movement to %s", motor.getName(), pos)
        motion = self.getMotion(motors)
        state, pos = motion.move(positions)
        if state != DevState.ON:
            self.warning("Motion ended in %s", state)
            msg = []
            for motor in motors:
                msg.append(motor.information())
            self.info("\n".join(msg))

class mstate(Macro):
    """Prints the state of a motor"""
    
    param_def = [['motor', Type.Moveable, None, 'Motor to check state']]

    def run(self, motor):
        self.info("Motor %s" % str(motor.getState()))

class umv(Macro):
    """Move motor(s) to the specified position(s) and update"""

    param_def = mv.param_def

    def prepare(self, *motor_pos_list, **opts):
        self.all_names = []
        self.all_pos = []
        self.print_pos = False
        for motor, pos in motor_pos_list:
            self.all_names.append([motor.getName()])
            pos, posObj = motor.getPosition(force=True), motor.getPositionObj()
            self.all_pos.append([pos])
            posObj.subscribeEvent(self.positionChanged, motor)
        
    def run(self, *motor_pos_list):
        self.print_pos = True
        pars = []
        for motor, pos in motor_pos_list:
            pars.append(motor)
            pars.append(pos)

        self.execMacro('mv', *pars)
        self.finish()
 
    def finish(self):
        self._clean()
        self.printAllPos()
 
    def on_abort(self):
        self.finish()
        
    def _clean(self):
        for motor, pos in self.getParameters():
            posObj = motor.getPositionObj()
            try:
                posObj.unsubscribeEvent(self.positionChanged, motor)
            except Exception, e:
                print str(e)
                raise e
    
    def positionChanged(self, motor, position):
        idx = self.all_names.index([motor.getName()])
        self.all_pos[idx] = [position]
        if self.print_pos:
            self.printAllPos()
    
    def printAllPos(self):
        motor_width = 10
        table = Table(self.all_pos, elem_fmt=['%*.4f'],
                      col_head_str=self.all_names, col_head_width=motor_width)
        self.outputBlock(table.genOutput())
        self.flushOutput()

class mvr(Macro):
    """Move motor(s) relative to the current position(s)"""

    param_def = [
       ['motor_disp_list',
        ParamRepeat(['motor', Type.Moveable, None, 'Motor to move'],
                    ['disp',  Type.Float, None, 'Relative displacement']),
        None, 'List of motor/displacement pairs'],
    ]
    
    def run(self, *motor_disp_list):
        motor_pos_list = []
        for motor, disp in motor_disp_list:
            pos = motor.getPosition(force=True) 
            if pos is None:
                self.error("Cannot get %s position" % motor.getName())
                return
            else:
                pos += disp
            motor_pos_list.extend([motor, pos])
        self.execMacro('mv', *motor_pos_list)

class umvr(Macro):
    """Move motor(s) relative to the current position(s) and update"""

    param_def = mvr.param_def

    def run(self, *motor_disp_list):
        motor_pos_list = []
        for motor, disp in motor_disp_list:
            pos = motor.getPosition(force=True) 
            if pos is None:
                self.error("Cannot get %s position" % motor.getName())
                return
            else:
                pos += disp
            motor_pos_list.extend([motor, pos])
        self.execMacro('umv', *motor_pos_list)


################################################################################
#
# Data acquisition related macros
#
################################################################################


class ct(Macro):
    """Count for the specified time on the active measurement group"""

    env = ('ActiveMntGrp',)

    param_def = [
       ['integ_time', Type.Float, 1.0, 'Integration time']
    ]

    def prepare(self, integ_time, **opts):
        mnt_grp_name = self.getEnv('ActiveMntGrp')
        self.mnt_grp = self.getObj(mnt_grp_name, type_class=Type.MeasurementGroup)

    def run(self, integ_time):
        self.output("hello")
        if self.mnt_grp is None:
            self.error('ActiveMntGrp is not defined or has invalid value')
            return

        self.debug("Counting for %s sec", integ_time)
        self.outputDate()
        self.output('')
        self.flushOutput()

        state, data = self.mnt_grp.count(integ_time)

        names, counts = [], []
        for ch_info in self.mnt_grp.getChannelsInfo():
            names.append('  %s' % ch_info.label)
            ch_data = data.get(ch_info.full_name)
            if ch_info.shape > [1]:
                counts.append(list(ch_data.shape))
            else:
                counts.append(ch_data)
        
        table = Table([counts], row_head_str=names, row_head_fmt='%*s',
                      col_sep='  =  ')
        for line in table.genOutput():
            self.output(line)


class uct(Macro):
    """Count on the active measurement group and update"""

    env = ('ActiveMntGrp',)

    param_def = [
       ['integ_time', Type.Float, 1.0, 'Integration time']
    ]

    def prepare(self, integ_time, **opts):
        
        self.print_value = False
        
        mnt_grp_name = self.getEnv('ActiveMntGrp')
        self.mnt_grp = self.getObj(mnt_grp_name, type_class=Type.MeasurementGroup)

        if self.mnt_grp is None:
            return

        names, nan = self.mnt_grp.getChannelLabels(), float('nan')
        self.names    = [ [n] for n in names ]
        
        self.values   = len(names)*[ [nan] ]
        self.channels = self.mnt_grp.getChannelAttrExs()

        for ch_attr_ex in self.channels:
            ch_attr_ex.subscribeEvent(self.counterChanged, ch_attr_ex)

    def printAllValues(self):
        ch_width = 10
        table = Table(self.values, elem_fmt=['%*.4f'], col_head_str=self.names, 
                      col_head_width=ch_width)
        self.outputBlock(table.genOutput())
        self.flushOutput()
    
    def counterChanged(self, ch_attr, value):
        idx = self.channels.index(ch_attr)
        self.values[idx] = [value]
        if self.print_value:
            self.printAllValues()
        
    def run(self, integ_time):
        if self.mnt_grp is None:
            self.error('ActiveMntGrp is not defined or has invalid value')
            return
        
        self.print_value = True
        
        state, data = self.mnt_grp.count(integ_time)

        for ch_attr_ex in self.mnt_grp.getChannelAttrExs():
            ch_attr_ex.unsubscribeEvent(self.counterChanged, ch_attr_ex)
        self.printAllValues()


class settimer(Macro):
    """Defines the timer channel for the active measurement group"""
    
    env = ('ActiveMntGrp',)
    
    param_def = [
       ['timer', Type.ExpChannel,   None, 'Timer'],
    ]
    
    def run(self,timer):
        mnt_grp_name = self.getEnv('ActiveMntGrp')
        mnt_grp = self.getObj(mnt_grp_name, type_class=Type.MeasurementGroup)

        if mnt_grp is None:
            self.error('ActiveMntGrp is not defined or has invalid value.\n' \
                       'please define a valid active measurement group ' \
                       'before setting a timer')
            return
        
        try:
            mnt_grp.setTimer(timer.getName())
        except Exception,e:
            self.output(str(e))
            self.output("%s is not a valid channel in the active measurement group" % timer)

@macro([['message', ParamRepeat(['message_item', Type.String, None, 'message item to be reported']), None, 'message to be reported']])
def report(self, *message):
    """Logs a new record into the message report system (if active)"""
    self.report(' '.join(message))



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""The MacroServer tango module"""

import os.path

from PyTango import Util, Except, DevVoid, DevLong, DevString, DevState, \
    DevEncoded, DevVarStringArray, READ, READ_WRITE, SCALAR, SPECTRUM, DebugIt

#from taurus.core.util import Logger
from taurus.core.util.codecs import CodecFactory

from sardana import State, SardanaServer #, ElementType
from sardana.tango.core.SardanaDevice import SardanaDevice, SardanaDeviceClass
from sardana.macroserver.msexception import MacroServerException
from sardana.macroserver.macroserver import MacroServer as MS


class MacroServer(SardanaDevice):
    """The MacroServer tango class"""
    
    ElementsCache = None
    EnvironmentCache = None
    
    def __init__(self,cl, name):
        self._macro_server = None
        SardanaDevice.__init__(self,cl, name)
    
    def init(self, name):
        SardanaDevice.init(self, name)
        
        if self._alias is None:
            self._alias = Util.instance().get_ds_inst_name()
        
        self._macro_server = ms = MS(self.get_full_name(), self.alias)
        ms.add_listener(self.on_macro_server_changed)
    
    @property
    def macro_server(self):
        return self._macro_server
    
    def delete_device(self):
        SardanaDevice.delete_device(self)
        self.clear_log_report()

    def init_device(self):
        SardanaDevice.init_device(self)
        self.set_change_event('State', True, False)
        self.set_change_event('Status', True, False)
        self.set_change_event('TypeList', True, False)
        self.set_change_event('DoorList', True, False)
        self.set_change_event('MacroList', True, False)
        self.set_change_event('MacroLibList', True, False)
        self.set_change_event('Elements', True, False)
        self.set_change_event('Environment', True, False)
        
        dev_class = self.get_device_class()
        self.get_device_properties(dev_class)
        
        self.EnvironmentDb = self._calculate_name(self.EnvironmentDb)
        self.LogReportFilename = self._calculate_name(self.LogReportFilename)

        macro_server = self.macro_server
        macro_server.set_python_path(self.PythonPath)
        macro_server.set_max_parallel_macros(self.MaxParallelMacros)
        
        # if it is not possible to store/retrieve the environment from the 
        # current path then setup a new unique path and store the environment 
        # there forever
        try:
            macro_server.set_environment_db(self.EnvironmentDb)
        except:
            self.error("Failed to set environment DB to %s", self.EnvironmentDb)
            self.debug("Details:", exc_info=1)
            import tempfile
            env_db = os.path.join(tempfile.mkdtemp(),
                                  MacroServerClass.DefaultEnvRelDir)
            env_db = self._calculate_name(env_db)
            db = Util.instance().get_database()
            db.put_device_property(self.get_name(), dict(EnvironmentDb=env_db))
            self.EnvironmentDb = env_db
            macro_server.set_environment_db(self.EnvironmentDb)
        
        try:
            macro_server.set_log_report(self.LogReportFilename, self.LogReportFormat)
        except:
            self.error("Failed to setup log report to %s",
                       self.LogReportFilename)
            self.debug("Details:", exc_info=1)
        
        macro_server.set_macro_path(self.MacroPath)
        macro_server.set_pool_names(self.PoolNames)
        
        if self.RConsolePort:
            try:
                import rfoo.utils.rconsole
                rfoo.utils.rconsole.spawn_server(port=self.RConsolePort)
            except Exception:
                self.warning("Failed to start rconsole")
                self.debug("Details:", exc_info=1)
        self.set_state(DevState.ON)
    
    def _calculate_name(self, name):
        if name is None:
            return None
        util = Util.instance()
        return name % { 'ds_name' : util.get_ds_name().lower(),
                        'ds_exec_name' : util.get_ds_exec_name(),
                        'ds_inst_name' : util.get_ds_inst_name().lower() }
    
    def on_macro_server_changed(self, evt_src, evt_type, evt_value):
        # during server startup and shutdown avoid processing element
        # creation events
        if SardanaServer.server_state != State.Running:
            return
        
        evt_name = evt_type.name.lower()
        
        multi_attr = self.get_device_attr()
        elems_attr = multi_attr.get_attr_by_name("Elements")
        if evt_name == "poolelementschanged":
            # force the element list cache to be rebuild next time someone reads
            # the element list
            self.ElementsCache = None
            self.set_attribute(elems_attr, value=evt_value.value)
            #self.push_change_event('Elements', *evt_value.value)
        elif evt_name in ("elementcreated", "elementdeleted"):
            # force the element list cache to be rebuild next time someone reads
            # the element list
            self.ElementsCache = None

            elem = evt_value
            
            value = { }
            if "created" in evt_name:
                key = 'new'
            else:
                key = 'del'
            json_elem = elem.serialize(pool=self.pool.full_name)
            value[key] = json_elem,
            value = CodecFactory().getCodec('json').encode(('', value))
            self.set_attribute(elems_attr, value=value)
            #self.push_change_event('Elements', *value)
        elif evt_name == "elementschanged":
            # force the element list cache to be rebuild next time someone reads
            # the element list
            self.ElementsCache = None
            
            ms_name = self.macro_server.full_name
            new_values, changed_values, deleted_values = [], [], []
            for elem in evt_value['new']:
                json_elem = elem.serialize(macro_server=ms_name)
                new_values.append(json_elem)
            for elem in evt_value['change']:
                json_elem = elem.serialize(macro_server=ms_name)
                changed_values.append(json_elem)
            for elem in evt_value['del']:
                json_elem = elem.serialize(macro_server=ms_name)
                deleted_values.append(json_elem)
            value = { "new" : new_values, "change": changed_values,
                      "del" : deleted_values }
            value = CodecFactory().getCodec('json').encode(('', value))
            self.set_attribute(elems_attr, value=value)
            #self.push_change_event('Elements', *value)
        elif evt_name == "environmentchanged":
            self.EnvironmentCache = None
            env_attr = multi_attr.get_attr_by_name("Environment")
            value = CodecFactory().getCodec('pickle').encode(('', evt_value))
            self.set_attribute(env_attr, value=value)
            
    
    def always_executed_hook(self):
        pass
    
    def read_attr_hardware(self,data):
        pass
    
    def read_DoorList(self, attr):
        door_names = self.macro_server.get_door_names()
        attr.set_value(door_names)
        
    @DebugIt()
    def read_MacroList(self, attr):
        macro_names = self.macro_server.get_macro_names()
        attr.set_value(macro_names)
    
    def read_MacroLibList(self, attr):
        macro_lib_names = self.macro_server.get_macro_lib_names()
        attr.set_value(macro_lib_names)
    
    def read_TypeList(self, attr):
        type_names = self.macro_server.get_data_type_names_with_asterisc()
        attr.set_value(type_names)
        
    #@DebugIt()
    def getElements(self, cache=True):
        value = self.ElementsCache
        if cache and value is not None:
            return value
        elements = self.macro_server.get_elements_info()
        value = dict(new=elements)
        value = CodecFactory().getCodec('json').encode(('', value))
        self.ElementsCache = value
        return value
    
    #@DebugIt()
    def read_Elements(self, attr):
        fmt, data = self.getElements()
        attr.set_value(fmt, data)

    def is_Elements_allowed(self, req_type):
        return SardanaServer.server_state == State.Running

    is_DoorList_allowed = \
    is_MacroList_allowed = \
    is_MacroLibList_allowed = \
    is_TypeList_allowed = is_Elements_allowed
        
    def GetMacroInfo(self, macro_names):
        """GetMacroInfo(list<string> macro_names):
        
           Returns a list of string containing macro information.
           Each string is a JSON encoded.
           
           Params:
               - macro_name: a list of strings with the macro(s) name(s)
           Returns:
               - a list of string containing macro information.
        """
        macro_server = self.macro_server
        codec = CodecFactory().getCodec('json')
        ret = [ codec.encode(('', macro.serialize()))[1]
            for macro in macro_server.get_macros()
                if macro.name in macro_names ]
        return ret
    
    def ReloadMacro(self, macro_names):
        """ReloadMacro(list<string> macro_names):"""
        try:
            for macro_name in macro_names:
                self.macro_server.reload_macro(macro_name)
        except MacroServerException, mse:
            Except.throw_exception(mse.type, mse.msg, 'ReloadMacro')
        return ['OK']
    
    def ReloadMacroLib(self, lib_names):
        """ReloadMacroLib(sequence<string> lib_names):
        """
        try:
            for lib_name in lib_names:
                self.macro_server.reload_macro_lib(lib_name)
        except MacroServerException, mse:
            Except.throw_exception(mse.type, mse.msg, 'ReloadMacroLib')
        return ['OK']
    
    def GetMacroCode(self, argin):
        """GetMacroCode(<module name> [, <macro name>]) -> full filename, code, line_nb
        """
        ret = self.macro_server.get_or_create_macro_lib(*argin)
        return map(str, ret)
    
    def SetMacroCode(self, argin):
        lib_name, code = argin[:2]
        auto_reload = True
        if len(argin) > 2:
            auto_reload = argin[2].lower() in ('true', 'yes')
        self.macro_server.set_macro_lib(lib_name, code, auto_reload=auto_reload)

    #@DebugIt()
    def getEnvironment(self, cache=True):
        value = self.EnvironmentCache
        if cache and value is not None:
            return value
        env = self.macro_server.get_env()
        value = dict(new=env)
        value = CodecFactory().getCodec('pickle').encode(('', value))
        self.EnvironmentCache = value
        return value
    
    def read_Environment(self, attr):
        fmt, data = self.getEnvironment()
        attr.set_value(fmt, data)
    
    def write_Environment(self, attr):
        data = attr.get_write_value()
        data = CodecFactory().getCodec('pickle').decode(data)[1]
        self.macro_server.change_env(data)
    
    def is_Environment_allowed(self, req_type):
        return True


class MacroServerClass(SardanaDeviceClass):
    """MacroServer Tango class class"""
    
    #    Class Properties
    class_property_list = {
    }
    
    
    DefaultEnvBaseDir = "/tmp/tango"
    DefaultEnvRelDir = "%(ds_exec_name)s/%(ds_inst_name)s/macroserver.properties"
    
    DefaultLogReportFormat = '%(levelname)-8s %(asctime)s: %(message)s'
    
    #    Device Properties
    device_property_list = {
        'PoolNames':
            [DevVarStringArray,
            "Sardana device pool device names",
            [] ],
        'MacroPath':
            [DevVarStringArray,
            "list of directories to search for macros (path separators "
            "can be '\n' or ':')",
            [] ],
        'PythonPath':
            [DevVarStringArray,
            "list of directories to be appended to sys.path at startup (path "
            "separators can be '\n' or ':')",
            [] ],
        'MaxParallelMacros':
            [DevLong,
            "Maximum number of macros that can execute concurrently.",
            [10] ],
        'EnvironmentDb':
            [DevString,
            "The environment database (usually a plain file).",
            os.path.join(DefaultEnvBaseDir, DefaultEnvRelDir) ],
        'RConsolePort':
            [DevLong,
            "The rconsole port number",
            None ],
        'LogReportFilename':
            [DevString,
            "Filename (absolute) which contains user log reports [default: "
            "None, meaning don't store log report messages]. The system will "
            "save old log files by appending extensions to the filename. The "
            "extensions are date-and-time based, using the strftime "
            "format %Y-%m-%d_%H-%M-%S or a leading portion thereof, "
            "depending on the rollover interval.",
            None ],
        'LogReportFormat':
            [DevString,
            "Log report format [default: '%s']" % DefaultLogReportFormat,
            DefaultLogReportFormat],
    }

    #    Command definitions
    cmd_list = {
        'GetMacroInfo':
            [[DevVarStringArray, "Macro(s) name(s)"],
            [DevVarStringArray, "Macro(s) description(s)"]],
        'ReloadMacro':
            [[DevVarStringArray, "Macro(s) name(s)"],
            [DevVarStringArray, "[OK] if successfull or a traceback " \
                "if there was a error (one string with complete traceback of " \
                "each error)"]],
        'ReloadMacroLib':
            [[DevVarStringArray, "MacroLib(s) name(s)"],
            [DevVarStringArray, "[OK] if successfull or a traceback " \
                "if there was a error (one string with complete traceback of " \
                "each error)" ]],
        'GetMacroCode':
            [[DevVarStringArray, "<MacroLib name> [, <Macro name>]"],
            [DevVarStringArray, "result is a sequence of 3 strings:\n"
                "<full path and file name>, <code>, <line number>" ]],
        'SetMacroCode':
            [[DevVarStringArray, "<MacroLib name>, <code> [, <Auto reload>=True]\n" \
                "- if macro lib is a simple module name:\n" \
                "  - if it exists, it is overwritten, otherwise a new python " \
                "file is created in the directory of the first element in "\
                "the MacroPath property" \
                "- if macro lib is the full path name:\n" \
                "  - if path is not in the MacroPath, an exception is thrown" \
                "  - if file exists it is overwritten otherwise a new file " \
                "is created"],
            [DevVoid, "" ]],
    }

    #    Attribute definitions
    attr_list = {
        'DoorList'     : [ [ DevString,  SPECTRUM, READ, 256 ] ],
        'MacroList'    : [ [ DevString,  SPECTRUM, READ, 4096 ] ],
        'MacroLibList' : [ [ DevString,  SPECTRUM, READ, 1024 ] ],
        'TypeList'     : [ [ DevString,  SPECTRUM, READ, 256 ] ],
        'Elements'     : [ [ DevEncoded, SCALAR,   READ ],
                           { 'label'       : "Elements",
                             'description' : "the list of all elements "
                                             "(a JSON encoded dict)", } ],
        'Environment'  : [ [ DevEncoded, SCALAR, READ_WRITE],
                           { 'label'       : 'Environment',
                             'description' : "The macro server environment "
                                             "(a JSON encoded dict)", } ],
    }
    



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains the definition of a discrete pseudo motor controller
for the Sardana Device Pool"""

__all__ = ["DiscretePseudoMotorController"]

__docformat__ = 'restructuredtext'

import json

from sardana import DataAccess
from sardana.pool.controller import PseudoMotorController
from sardana.pool.controller import Type, Access, Description

CALIBRATION = 'Calibration'
LABELS = 'Labels'


class DiscretePseudoMotorController(PseudoMotorController):
    """A discrete pseudo motor controller which converts physical motor positions
    to discrete values"""
    
    gender = "DiscretePseudoMotorController"
    model  = "PseudoMotor"
    organization = "Sardana team"
    image = ""

    pseudo_motor_roles = ("OutputMotor",)
    motor_roles = ("InputMotor",)

    axis_attributes = {CALIBRATION:#type hackish until arrays supported
                        {Type : str,
                         Description : 'Flatten list of a list of triples and [min,cal,max]',
                         Access : DataAccess.ReadWrite,
                         'fget' : 'get%s' % CALIBRATION,
                         'fset' : 'set%s' % CALIBRATION},
                       LABELS:#type hackish until arrays supported
                        {Type : str,
                         Description : 'String list with the meaning of each discrete position',
                         Access : DataAccess.ReadWrite,
                         'fget' : 'get%s' % LABELS,
                         'fset' : 'set%s' % LABELS}
                      }


    def __init__(self, inst, props, *args, **kwargs):
        PseudoMotorController.__init__(self, inst, props, *args, **kwargs)
        self._calibration = None
        self._positions = None
        self._labels = None


    def GetAxisAttributes(self, axis):
        axis_attrs = PseudoMotorController.GetAxisAttributes(self, axis)
        axis_attrs = dict(axis_attrs)
        axis_attrs['Position']['type'] = int
        return axis_attrs


    def CalcPseudo(self, axis, physical_pos, curr_pseudo_pos):
        llabels = len(self._labels)
        positions = self._positions
        calibration = self._calibration
        lcalibration = len(calibration)

        value = physical_pos[0]
        #case 0: nothing to translate, only round about integer the attribute value
        if llabels == 0:
            return int(value)
        #case 1: only uses the labels. Available positions in POSITIONS
        elif lcalibration == 0:
            value = int(value)
            try: positions.index(value)
            except: raise Exception("Invalid position.")
            else: return value
        #case 1+fussy: the physical position must be in one of the defined
        #ranges, and the DiscretePseudoMotor position is defined in labels
        elif llabels == lcalibration:
            for fussyPos in calibration:
                if value >= fussyPos[0] and value <= fussyPos[2]:
                    return positions[calibration.index(fussyPos)]
            #if the loop ends, current value is not in the fussy areas.
            raise Exception("Invalid position.")
        else:
            raise Exception("Bad configuration on axis attributes.")


    def CalcPhysical(self, axis, pseudo_pos, curr_physical_pos):
        #If Labels is well defined, the write value must be one this struct
        llabels = len(self._labels)
        positions = self._positions
        calibration = self._calibration
        lcalibration = len(calibration)
        value = pseudo_pos[0]

        #case 0: nothing to translate, what is written goes to the attribute
        if llabels == 0:
            return value
        #case 1: only uses the labels. Available positions in POSITIONS
        elif lcalibration == 0:
            self._log.debug("Value = %s", value)
            try: positions.index(value)
            except: raise Exception("Invalid position.")
            return value
        #case 1+fussy: the write to the to the DiscretePseudoMotorController
        #is translated to the central position of the calibration.
        elif llabels == lcalibration:
            self._log.debug("Value = %s", value)
            try: destination = positions.index(value)
            except: raise Exception("Invalid position.")
            self._log.debug("destination = %s", destination)
            calibrated_position = calibration[destination][1]#central element
            self._log.debug("calibrated_position = %s", calibrated_position)
            return calibrated_position


    def getLabels(self,axis):
        #hackish until we support DevVarDoubleArray in extra attrs
        labels = self._labels
        positions = self._positions
        labels_str = ""
        for i in range(len(labels)):
            labels_str += "%s:%d "%(labels[i],positions[i])
        return labels_str[:-1]#remove the final space


    def setLabels(self,axis,value):
        #hackish until we support DevVarStringArray in extra attrs
        labels = []
        positions = []
        for pair in value.split():
            l,p = pair.split(':')
            labels.append(l)
            positions.append(int(p))
        if len(labels) == len(positions):
            self._labels = labels
            self._positions = positions
        else:
            raise Exception("Rejecting labels: invalid structure")


    def getCalibration(self,axis):
        return json.dumps(self._calibration)


    def setCalibration(self,axis,value):
        try:
            self._calibration = json.loads(value)
        except:
            raise Exception("Rejecting calibration: invalid structure")
